#!/usr/bin/env python
#
# Author: Thamme Gowda [tg (at) isi (dot) edu] 
# Created: 2019-10-29
import argparse
from typing import Dict, Any, Iterator, TextIO
import sys
from pathlib import Path

from nlcodec import learn_vocab, load_scheme, encode, decode, log
from nlcodec import __version__, __description__, __epilog__
from nlcodec import DEF_WORD_MIN_FREQ, DEF_CHAR_MIN_FREQ, DEF_CHAR_COVERAGE, DEF_MIN_CO_EV


def write_lines(lines: Iterator[str], out: TextIO, line_break='\n'):
    for line in lines:
        out.write(line)
        out.write(line_break)


class MyFormatter(argparse.ArgumentDefaultsHelpFormatter):

    def _split_lines(self, text, width: int):
        if text.startswith("R|"):
            return text[2:].splitlines()
        return super()._split_lines(text, width)


def parse_args() -> Dict[str, Any]:
    p = argparse.ArgumentParser(formatter_class=MyFormatter, prog='nlcodec',
                                description=__description__,
                                epilog=__epilog__)
    p.add_argument('-v', '--version', action='version', version=f'%(prog)s {__version__}')
    p.add_argument("task", choices=['learn', 'encode', 'decode', 'estimate'],
                   help='''R|"task" or sub-command.
    "learn" - learns vocabulary. use --level and vocab_size for type and size 
    "encode" - encodes a dataset 
    "decode" - decodes an already encoded dataset
    "estimate" - estimates quality attributes of an encoding''')

    p.add_argument('-i', '--inp', type=argparse.FileType('r'), default=sys.stdin,
                   help='Input file path')
    p.add_argument('-tfs', '--term-freqs', action='store_true', default=False,
                   help='--inp is term frequencies Valid for task=learn.'
                        ' See nlcodec.term_freq to obtain them')
    p.add_argument('-o', '--out', type=argparse.FileType('w'), default=sys.stdout,
                   help='Output file path. Not valid for "learn" or "estimate" task')
    p.add_argument('-m', '--model', type=Path, help='Path to model aka vocabulary file',
                   required=True)

    p.add_argument('-idx', '--indices', action='store_true', default=None,
                   help='Indices instead of strings. Valid for task=encode and task=decode')

    learn_args = p.add_argument_group("args for task=learn")
    learn_args.add_argument('-vs', '--vocab-size', type=int, default=-1,
                            help='Vocabulary size. Valid only for task=learn. This is required for'
                                 ' "bpe", but optional for "word" and "char" models, specifying it'
                                 ' will trim the vocabulary at given top most frequent types.')
    learn_args.add_argument('-l', '--level', choices=['char', 'word', 'bpe'],
                            help='Encoding Level; Valid only for task=learn')
    learn_args.add_argument('-mf', '--min-freq', default=None, type=int,
                            help='Minimum frequency of types for considering inclusion in vocabulary. '
                            'Types fewer than this frequency will be ignored. '
                            f'For --level=word or --level=bpe, freq is type freq and '
                            f' default is {DEF_WORD_MIN_FREQ}.'
                            f'for --level=char, characters fewer than this value'
                            f' will be excluded. default={DEF_CHAR_MIN_FREQ}')

    learn_args.add_argument('-cv', '--char-coverage', default=DEF_CHAR_COVERAGE, type=float,
                            help='Character coverage for --level=char or --level=bpe')

    learn_args.add_argument('-mce', '--min-co-ev', default=DEF_MIN_CO_EV, type=int,
                            help='Minimum Co-evidence for BPE merge.'
                                 ' Valid when --task=learn and --level=bpe')

    args = vars(p.parse_args())
    return args


def main():
    args = parse_args()
    task = args.pop('task')
    if task == 'learn':
        args.pop('out')  # No output
        args.pop('indices')  # No output
        assert args.get('level'), 'argument --level is required for "learn" task'
        import time
        from datetime import timedelta
        from nlcodec.utils import max_RSS
        st = time.time()
        st_mem = max_RSS()[1]
        learn_vocab(**args)
        delta = timedelta(seconds=time.time() - st)
        et_mem = max_RSS()[1]
        log.info(f"Time taken: {delta}; Memory: {st_mem} --> {et_mem}")
    elif task in ('encode', 'decode'):
        scheme = load_scheme(args.pop('model'))
        inp, out, indices = args['inp'], args['out'], args.get('indices', False)
        if task == 'encode':
            recs = encode(inp, scheme, indices=indices)
            if indices:
                recs = ([str(idx) for idx in seq] for seq in recs)
            recs = (' '.join(seq) for seq in recs)
        else:
            recs = decode(inp, scheme, indices=indices)
        write_lines(recs, out)
    elif task == 'estimate':
        from nlcodec.qestim import estimate
        estimate(codec_path=args['model'], data=args['inp'])
    else:
        raise NotImplementedError(task + ' not implemented')


if __name__ == '__main__':
    main()



#!/usr/bin/env python
#
# Author: Thamme Gowda [tg (at) isi (dot) edu] 
# Created: 2019-10-25

import collections as coll
import copy
import time
from typing import List, Dict, Tuple, Union, Iterator, Set

from nlcodec import log, DEF_MIN_CO_EV
from nlcodec.codec import Type, Level, Reseved
from nlcodec.dstruct import LnNode, MaxHeap
from nlcodec.utils import max_RSS
from tqdm import tqdm

Codes = Dict[int, Tuple[int, ...]]
Seq = List[int]
Bigram = Tuple[int, int]


class BPELearn:
    """
    The core BPE learning algorithm
    fast implementation using linked lists
    Note: this implementation takes relatively more RAM; and that is okay for my usecase
    # TODO: write this in c++ or rust and bind it here
    """
    space_tok = Reseved.SPACE_TOK[0]
    unk_tok = Reseved.UNK_TOK[0]

    def __init__(self, seqs: Iterator[Union[Seq, Tuple[Seq, int]]], vocab: List[Type]):
        # Check one to one map: type.name <-> idx
        assert len(set(v.idx for v in vocab)) == len(set(v.name for v in vocab))
        for i, v in enumerate(vocab):
            assert i == v.idx
        self.vocab = vocab
        self.uni: Dict[int, int] = coll.defaultdict(int)  # term freq ; unigrams
        self.bi: Dict[Bigram, int] = coll.defaultdict(int)  # bigram frequencies

        # Bigram to sequence references
        self.bi_ixs: Dict[Bigram, Set[LnNode]] = coll.defaultdict(set)

        self.create_index(seqs)
        self.validate_index()

    def create_index(self, seqs):
        log.info("Going to build corpus stats index; This might take lot of time and memory")
        n_seqs, n_ignored, n_replaced, bar_msg = 0, 0, 0, ''
        with tqdm(enumerate(seqs), unit='seqs', dynamic_ncols=True, mininterval=1) as data_bar:
            for idx, seq in data_bar:
                freq = 1  # default = 1 freq
                if isinstance(seq, tuple):  # if freq is available
                    seq, freq = seq

                n_seqs += 1
                if idx == 0:  # basic validation
                    assert isinstance(seq, list)  # first sequence, tokenized
                    assert isinstance(seq[0], int)  # sequence's item, should be an int or codepoint
                if not seq:
                    log.warning(f"Skipping empty sequence at idx {idx + 1}")
                    continue

                nodes = LnNode.from_seq(seq, freq=freq)
                assert len(seq) == len(nodes)

                for i in range(len(seq) - 1):  # the last position left out
                    bigm = (seq[i], seq[i + 1])
                    self.bi[bigm] += freq
                    assert nodes[i] not in self.bi_ixs[bigm]
                    self.bi_ixs[bigm].add(nodes[i])  # bigm found at node i
                    self.uni[seq[i]] += freq
                self.uni[seq[-1]] += freq  # the last unigram count; not covered in the above loop
                bar_msg = f'MaxRSS={max_RSS()[1]}'
                data_bar.set_postfix_str(bar_msg, refresh=False)
        log.info(f"Created index; {bar_msg}")

    @property
    def vocab_size(self) -> int:
        return len(self.vocab)

    def validate_index(self):
        """
        Call this any time to check if the index of uni bi bi_ixs are valid.
        Raises exception on invalid index
        :return:
        """
        # This is code doesnt work with fast but new dirty heap updates
        max_code = max(self.uni)
        max_idx = max(t.idx for t in self.vocab)
        if not (max_code < self.vocab_size and max_code <= max_idx):
            raise ValueError(
                f'Vocab size is {self.vocab_size}, but max_code is {max_code}; max_idx={max_idx}')
        if not len(self.bi) == len(self.bi_ixs):
            raise ValueError(f"|bi|={len(self.bi)} and |bi_idxs|={len(self.bi_ixs)} are not same")
        for bigm, freq in self.bi.items():
            if not freq >= 0:
                raise ValueError(f"{bigm} has freq {freq}; expected positive val")
            if not bigm in self.bi_ixs:
                raise ValueError(f"{bigm} exists in bi but not in bi_ixs")
            idx_freq = sum(n.freq for n in self.bi_ixs[bigm])
            if not freq == idx_freq:
                raise ValueError(
                    f"{bigm} has freq={freq} bi but has {idx_freq} bi_ixs refs")
            # less than unigram freqs
            if not freq <= self.uni[bigm[0]]:
                raise ValueError(f"{bigm} has freq={freq} bi but {bigm[0]} has {self.uni[bigm[0]]}")
            if not freq <= self.uni[bigm[1]]:
                raise ValueError(f"{bigm} has freq={freq} bi but {bigm[1]} has {self.uni[bigm[1]]}")
        for uni, freq in self.uni.items():
            if not freq >= 0:
                raise ValueError(f"{uni} has freq={freq}; expected positive value")
        log.info(f"Index is valid")

    def learn_codes(self, n_merges: int, min_co_evidence, code_level: int,
                    log_every=2) -> List[Type]:
        """
        :param n_merges: how many more merges
        :param min_co_evidence: min evidence (co-occurrence frequency);
         causes early stop upon failure
        :param code_level: what level to use for new code types created during merge
            for instance level=1 for word bpe; level=2 for seq bpe
        :param log_every: delay, in seconds between logs
        :return:
        """
        uni, bi_ixs = self.uni, self.bi_ixs
        heap = MaxHeap(self.bi)
        heap_dirty = coll.defaultdict(int)  # subtractions aren't updated in max-heap, they are here
        vocab = self.vocab
        last_log_t = time.time()
        log.info(f"logs every {log_every} seconds")
        for i in range(n_merges):
            # Using MaxHeap for faster lookup of max. But heap gets a bit dirty, so a bit of cleanup
            max_pair, pair_freq = heap.pop()
            while max_pair in heap_dirty:  # clean all max [airs until a clean value
                freq_update = heap_dirty.pop(max_pair)
                assert freq_update < 0  # only decrements are valid. increments make this wrong
                corr_freq = pair_freq + freq_update  # correct value
                assert corr_freq >= 0, f'{max_pair}:{pair_freq}, Δ={freq_update} = {corr_freq}'
                if corr_freq > 0:  # exclude zero count
                    heap.push(max_pair, corr_freq)
                max_pair, pair_freq = heap.pop()

            # here the  actual loop begins
            if pair_freq < min_co_evidence:
                log.warning(f"Early stop; max evidence found is {pair_freq} "
                            f"but min required is {min_co_evidence}")
                break

            new_type_idx = len(vocab)
            a, b = max_pair
            if time.time() - last_log_t >= log_every:
                log.info(f"{(100 * i / n_merges):.2f}% :: {new_type_idx} || {a:4}:{uni[a]:5}"
                         f" || {b:4}:{uni[b]:5} || {pair_freq:,} || {vocab[a].name} {vocab[b].name}")
                last_log_t = time.time()

            # code -> bigram   (flatten out bigram;  resolve interim codes
            new_type = Type(vocab[a].name + vocab[b].name, idx=new_type_idx, freq=pair_freq,
                            level=code_level, kids=(vocab[a], vocab[b]))
            vocab.append(new_type)

            # updates: update bigram and unigram counts
            uni[new_type_idx] = pair_freq  # this bigram is now a new unigram
            # unigram counts drop ; since some of their bigrams are removed
            uni[a] -= pair_freq
            uni[b] -= pair_freq
            heap_deltas = coll.defaultdict(int)
            update_nodes = bi_ixs.pop(max_pair)  # also removed from bi_ixs
            for node in update_nodes:
                # -- x a b y --
                x_node, b_node = node.left, node.right
                if node.is_unlinked or (a == b and new_type.idx in (node.val, b_node.val)):
                    # this happens in the cases like "x a a a a y"
                    uni[a] += node.freq
                    uni[b] += node.freq
                    uni[new_type.idx] -= node.freq
                    continue

                y_node = b_node.right
                dirty = node.val != a or b_node.val != b  # check that the linked list is proper
                if dirty:
                    log.warning(f'Expected {a, b} but found {node.val, b_node.val}'
                                f'\n {node, b_node}'
                                f'\n--{vocab[a].signature()} =='
                                f' {vocab[node.val].signature() if node.val != a else "OK"}'
                                f'\n--{vocab[b].signature()} =='
                                f' {vocab[b_node.val].signature() if b_node.val != b else "OK"}')
                    log.warning(f"a={a}, b={b} || a_node={node}, b_node={b_node}")
                assert not dirty
                assert node.freq == b_node.freq

                # update : x a b y => x R y
                b_node.delete(unlink=True)  # delete() takes care of linking a → y and a ← y
                new_node = node  # reuse a node as new_node/R
                new_node.val = new_type_idx  # reuse a as new_node/R
                # Note: the above edits to a and b nodes do-not/should-not change __hash__

                if x_node:
                    # remove (x_node_val, a) from bi and bi_ixs
                    heap_deltas[(x_node.val, a)] -= x_node.freq
                    if bi_ixs.get((x_node.val, a)):
                        # not sure why 'if' needed here;
                        bi_ixs[(x_node.val, a)].remove(x_node)

                    # add (x_node_val, R) to bi and bi_ixs
                    heap_deltas[(x_node.val, new_type_idx)] += x_node.freq
                    bi_ixs[(x_node.val, new_type_idx)].add(x_node)
                if y_node:
                    # remove (b, y_node.val) from bi and bi_ixs
                    heap_deltas[(b, y_node.val)] -= b_node.freq
                    if bi_ixs.get((b, y_node.val)):
                        # not sure why 'if' needed here;
                        bi_ixs[(b, y_node.val)].remove(b_node)

                    # add (R, y_node.val) to bi and bi_ixs
                    heap_deltas[(new_type_idx, y_node.val)] += b_node.freq
                    bi_ixs[(new_type_idx, y_node.val)].add(new_node)

            # however; the counts shouldn't go negative
            assert uni[a] >= 0
            assert uni[b] >= 0

            for pair, delta in heap_deltas.items():
                if delta > 0:  # these are new insertions, and they can go directly to heap
                    assert new_type_idx in pair
                    heap.push(pair, delta)
                elif delta < 0:  # one of those subtractions, which cant be directly updated
                    assert new_type_idx not in pair
                    heap_dirty[pair] += delta

        return vocab

    @classmethod
    def prepare_word(cls, word):
        # mark ending of sequences
        # # TODO: check:  looks like sentence piece adds at the beginning
        # subword-nmt (senrich et al 2016) did </w> at the end;
        # 0.2 of subword-nmt puts last char and </w> together
        return word + cls.space_tok

    @classmethod
    def _make_idxs(cls, voc_idx: Dict[str, int], term_freqs: Dict[str, int]) \
            -> Iterator[Tuple[Seq, int]]:
        """Convert character sequences to char indexed seqs"""
        unk_idx = voc_idx[cls.unk_tok]
        for word, freq in term_freqs.items():
            if word in voc_idx:
                res = [voc_idx[word]]
            else:
                res = [voc_idx.get(ch, unk_idx) for ch in word]
            yield res, freq

    @classmethod
    def _learn_codes(cls, term_freqs: Dict[str, int], vocab: List[Type], vocab_size: int,
                     init_list: List[str] = None,
                     min_co_evidence: int = DEF_MIN_CO_EV) -> List[Type]:
        """
        :param term_freqs: words types and frequencies
        :param vocab: initial vocab; usually reserved and alphabet
        :param vocab_size: desired vocabulary size
        :param init_list: any reserved words
        :return: List[str] word pieces
        """

        vocab = copy.copy(vocab)
        log.info(f"Found {len(term_freqs)} types before splitting; initial vocab {len(vocab)}")
        if init_list:
            log.info(f'Adding {len(init_list)} types to the initial vocab')
            assert not any(' ' in w for w in init_list), 'spaces not allowed in init_list words'
            vocab += [Type(cls.prepare_word(w), level=Level.user, idx=idx, freq=0)
                      for idx, w in enumerate(init_list, start=len(vocab))]

        rev_idx: Dict[str, int] = {word.name: word.idx for word in vocab}
        assert len(rev_idx) == len(vocab)  # one to one map
        assert vocab_size > len(vocab), f'vocab_size={vocab_size} is too small;' \
                                        f' found {len(vocab)} in the init vocab! Set a value larger than {len(vocab)}'

        seqs_freqs = cls._make_idxs(rev_idx, term_freqs)
        learner = BPELearn(seqs_freqs, vocab=vocab)
        final_vocab = learner.learn_codes(n_merges=vocab_size - len(vocab),
                                          min_co_evidence=min_co_evidence,
                                          code_level=Level.subword)
        return final_vocab

    @classmethod
    def learn_subwords(cls, term_freqs: Dict[str, int], vocab_size: int,
                       min_co_evidence: int = DEF_MIN_CO_EV,
                       init_vocab_factory=None) -> List[Type]:
        """
        :param term_freqs:
        :param vocab_size: final vocab size: reserved + chars + user_specified  + merges;
          special case, when `vocab_size=-1` the returned vocab will have just reserved + chars
        :param min_co_evidence: min co evidence for pair merges
        :param char_coverage: percentage of characters to be covered by inital char_freqs
        :param word_min_freq: words below this frequency will be excluded for learning BPE
        :return: List of Type
        """

        log.info(f"Total types: {len(term_freqs)}")
        term_freqs = {cls.prepare_word(word): freq for word, freq in term_freqs.items()}

        char_freqs = coll.defaultdict(int)
        for term, freq in term_freqs.items():
            for ch in term:
                char_freqs[ch] += freq
            """TODO: test this behavior; similar to subword-nmt v0.2
            for ch in term[:-2]:  # skip the last two: ending and the whitespace marker
                char_freqs[ch] += freq
            char_freqs[term[-2:]] += freq  # ending + whitespace marker go together as a single byte
            """
        init_vocab = init_vocab_factory(char_freqs)  # create an initial vocabulary of chars
        if vocab_size == -1:
            log.warning(f'Since vocab_size={vocab_size}; not going to do any L1 merges')
            log.info(f'Found initial vocab size of {len(init_vocab)}')
            return init_vocab

        return cls._learn_codes(term_freqs, init_vocab, min_co_evidence=min_co_evidence,
                                vocab_size=vocab_size)

    @classmethod
    def learn_subwords_from_corpus(cls, corpus: Iterator[str], **kwargs) -> List[Type]:
        """
        :param corpus: line iterator
        :param **kwargs : Refer learn_subwords() args
        """
        term_freq = coll.Counter(word for seq in corpus for word in seq.strip().split())
        return cls.learn_subwords(term_freq, **kwargs)



#!/usr/bin/env python
#
# Author: Thamme Gowda [tg (at) isi (dot) edu] 
# Created: 2019-11-12

import argparse
import sys
import logging as log
from pathlib import Path
from nlcodec import load_scheme, EncoderScheme
from typing import Iterator
from nlcodec.utils import make_n_grams_all
import json
from tqdm import tqdm

log.basicConfig(level=log.INFO)
debug_mode = log.getLogger().isEnabledFor(level=log.DEBUG)


def encode_all(inp : Iterator[str], scheme: EncoderScheme, index=True):
    """
    :param inp: iterator sentences
    :param scheme:  encoder scheme
    :param index: True to get integer indexes, False to get string pieces
    :return: Iterator of Seqs, where seq is encoder per scheme as index
    """
    for line in tqdm(inp, mininterval=1, dynamic_ncols=True):
        line = line.strip()
        if index:
            yield scheme.encode(line)
        else:
            yield scheme.encode_str(line)


def main(inp, out, n, model_path):
    scheme = load_scheme(model_path)
    seqs = scheme.encode_parallel(inp)
    freqs = make_n_grams_all(seqs, n)
    freqs = sorted(freqs.items(), key=lambda x:x[1], reverse=True)
    for gram, freq in freqs:
        gram = list(gram)
        names = [scheme.table[g].name for g in gram]
        line = json.dumps([gram, freq, names], ensure_ascii=False)
        out.write(line + '\n')


if __name__ == '__main__':
    p = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                description="Compute term frequencies for ")
    p.add_argument('-i', '--inp', type=argparse.FileType('r'), default=sys.stdin,
                   help='Input file path')
    p.add_argument('-o', '--out', type=argparse.FileType('w'), default=sys.stdout,
                   help='Output file path')
    p.add_argument('-m', '--model', dest='model_path', type=Path, required=True,
                   help="model aka vocabulary file")
    p.add_argument('-n', '--n', type=int, default=1, help='maximum n as in n-gram.')
    args = vars(p.parse_args())
    main(**args)



#!/usr/bin/env python
#
# Author: Thamme Gowda [tg (at) isi (dot) edu]
# Created: 2019-11-11

import argparse
import sys
import logging as log
import io
import collections as coll
from nlcodec.utils import make_n_grams, make_n_grams_all
from nlcodec import load_scheme, EncoderScheme
from pathlib import Path
import json

log.basicConfig(level=log.INFO)
debug_mode = log.getLogger().isEnabledFor(level=log.DEBUG)


def count_gram_recall(cands, refs, n=1):
    gram_recalls = coll.defaultdict(list)
    gram_precisions = coll.defaultdict(list)
    ref_gram_freqs = coll.defaultdict(int)
    cand_gram_freqs = coll.defaultdict(int)

    for cand, ref in zip(cands, refs):
        cand_grams = coll.Counter(make_n_grams(cand, n))
        ref_grams = coll.Counter(make_n_grams(ref, n))
        for ref_gram, ref_freq in ref_grams.items():
            assert ref_freq > 0
            cand_freq = cand_grams.get(ref_gram, 0)
            gram_recalls[ref_gram].append(min(cand_freq, ref_freq) / ref_freq)
            ref_gram_freqs[ref_gram] += ref_freq

        for cand_gram, cand_freq in cand_grams.items():
            assert cand_freq > 0
            ref_freq = ref_grams.get(cand_gram, 0)
            gram_precisions[cand_gram].append(min(cand_freq, ref_freq) / cand_freq)
            cand_gram_freqs[cand_gram] += cand_freq

    # average at the end; TODO: moving average
    gram_recalls = {gram: sum(recalls) / len(recalls) for gram, recalls in gram_recalls.items()}
    gram_precisions = {gram: sum(precs) / len(precs) for gram, precs in gram_precisions.items()}
    return gram_recalls, ref_gram_freqs, gram_precisions, cand_gram_freqs

def f1_measure(precison, recall):
    assert 0 <= precison <= 1
    assert 0 <= recall <= 1
    denr = precison + recall
    if denr == 0:
        return 0  # numerator 2*p*r is also zero
    return 2 * precison * recall / denr

def main(model_path, cands, refs, n, out, freqs=None):
    codec = load_scheme(model_path)
    cands, refs = list(cands), list(refs)
    assert len(cands) == len(refs), f'cands: {len(cands)} but refs: {len(refs)} lines'

    cands = list(codec.encode_parallel(cands))
    refs = list(codec.encode_parallel(refs))
    gram_recalls, ref_gram_freqs, gram_precisions, cand_gram_freqs = count_gram_recall(cands, refs)
    if freqs:
        log.info(f"Loading precomputed gram freqs from {freqs}")
        freqs = [json.loads(l.strip()) for l in freqs]
        gram_freqs = {tuple(g): f for g, f, name in freqs}

        # subset of grams that are found in reference
        gram_freqs = {g: f for g, f in gram_freqs.items() if g in ref_gram_freqs or g in cand_gram_freqs}

        # these grams were not found in training, but in there in refs => OOVs => freq=-1
        oov_grams = {g: -1 for g in ref_gram_freqs if g not in gram_freqs}
        log.info(f"{len(oov_grams)} grams were oov wrt to freqs => assigned freq = -1 ")
        gram_freqs.update(oov_grams)
    else:
        gram_freqs = ref_gram_freqs
    #print(gram_freqs.keys())
    new_grams = {cand_gram: freq for cand_gram, freq in cand_gram_freqs.items()
                   if cand_gram not in gram_freqs}

    if new_grams:
        msg = f'Found {len(new_grams)} grams that are not found in refs or --freqs'
        log.warning(msg)
        if n == 1:
            for ng, f in new_grams.items():
                ng = ng[0]
                log.error(f'Not found:\t{ng}\t{codec.idx_to_str[ng]}\t{f}')
            #raise Exception(msg)
        else:
            log.warning("TG, Come back and handle bigrams and above :)")

    gram_freqs = sorted(gram_freqs.items(), key=lambda t: t[1], reverse=True)
    out.write(f'Rank\tGram\tName\tRankF\tRefF\tCandF\tRecall\tPrecision\tF1\n')
    for i, (gram, rank_freq) in enumerate(gram_freqs):
        name = ','.join(codec.idx_to_str[g] for g in gram)
        idxs = ','.join(str(g) for g in gram)
        gram_recall = gram_recalls.get(gram, 0)
        gram_precision = gram_precisions.get(gram, 1) # should it be zero or one?
        f1 = f1_measure(gram_precision, gram_recall)
        ref_freq = ref_gram_freqs.get(gram, -1)
        cand_freq = cand_gram_freqs.get(gram, -1)
        out.write(f'{i+1}\t{idxs}\t{name}\t{rank_freq}\t{ref_freq}\t{cand_freq}\t{gram_recall:g}'
                  f'\t{gram_precision:g}\t{f1:g}\n')


def parse_args():
    stdin = io.TextIOWrapper(sys.stdin.buffer, encoding='utf-8', errors='ignore', newline='\n')
    stdout = io.TextIOWrapper(sys.stdout.buffer, encoding='utf-8', errors='ignore')
    p = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                description="Tool to compute Recall vs Frequency correlation.")
    p.add_argument('-c', '--cands', type=argparse.FileType('r', encoding='utf-8', errors='ignore'),
                   default=stdin, help='Candidate (aka output from NLG system) file')
    p.add_argument('-r', '--refs', type=argparse.FileType('r', encoding='utf-8', errors='ignore'),
                   default=stdin,
                   help='Reference (aka human label) file')
    p.add_argument('-f', '--freqs', type=argparse.FileType('r', encoding='utf-8', errors='ignore'),
                   help='precomputed freqs of grams on some other data (such as training) '
                        ' which should be used for ranking.'
                        ' If not given, --refs is used. This can be obtained from `termfreqs.py`')

    p.add_argument('-n', '--n', type=int, default=1, help='maximum n as in n-gram.')
    p.add_argument('-o', '--out', type=argparse.FileType('w'), default=stdout,
                   help='Output file path to store the result.')
    p.add_argument('-m', '--model', dest='model_path', type=Path, required=True,
                   help="model aka vocabulary file")
    args = vars(p.parse_args())
    assert not (args['cands'] == args['refs'] == stdin), \
        'Only one of --refs and --cands can be read from STDIN'
    return args


if __name__ == '__main__':
    args = parse_args()
    main(**args)



#!/usr/bin/env python
#
# Author: Thamme Gowda [tg (at) isi (dot) edu] 
# Created: 2019-10-16

import abc
import collections as coll
import json
from dataclasses import dataclass, fields
from datetime import datetime
from pathlib import Path
from typing import List, TextIO, Dict, Tuple, Union, Iterator, Optional
import multiprocessing as mp
from tqdm import tqdm
from nlcodec import __version__, log
from nlcodec.dstruct import TrNode
from nlcodec.utils import filter_types_coverage
import os
import sys
import random

N_CPUS = max(1, mp.cpu_count() - 1)
N_CPUS = int(os.environ.get('NLCODEC_THREADS', str(N_CPUS)))

assert N_CPUS >= 1
from nlcodec import DEF_WORD_MIN_FREQ as WORD_MIN_FREQ
from nlcodec import DEF_CHAR_MIN_FREQ as CHAR_MIN_FREQ
from nlcodec import DEF_CHAR_COVERAGE as CHAR_COVERAGE
from nlcodec import DEF_MIN_CO_EV as MIN_CO_EV


class Reseved:
    PAD_TOK = '<pad>', 0
    UNK_TOK = '<unk>', 1  # unk = '⁇'  # U+2047  to make up some OOV characters
    BOS_TOK = '<s>', 2
    EOS_TOK = '</s>', 3
    CLS_TOK = '<cls>', 4
    SPACE_TOK = '▁', 5  # U+2581 same as google/sentencepiece

    PAD_IDX = PAD_TOK[1]
    UNK_IDX = UNK_TOK[1]
    BOS_IDX = BOS_TOK[1]
    EOS_IDX = EOS_TOK[1]
    CLS_IDX = CLS_TOK[1]

    ALL = [PAD_TOK, UNK_TOK, BOS_TOK, EOS_TOK, CLS_TOK, SPACE_TOK]

    @classmethod
    def validate(cls, table: List['Type']):
        for tok, idx in cls.ALL:
            assert tok == table[idx].name, f'Cant find reserved {tok} at index {idx}'

    @classmethod
    def with_reserved_types(cls) -> List['Type']:
        return [Type(tok, level=Level.reserved, idx=idx, freq=-1) for tok, idx in cls.ALL]


class Level:
    reserved = -1
    char = 0
    user = 0  # as of now,  user and char are indistinguishable; 0 means dont look inside
    subword = 1
    word = 2
    phrase = 3


@dataclass(frozen=True)
class Type:  # Type as in word type vs token
    name: str  # name of the type
    level: int  # in [-1, 0, 1, 2, 3]
    idx: int  # idx for modeling integer sequence
    freq: int  # when available; from corpus
    kids: Optional[Tuple['Type', ...]] = None  # in case it was composed by sub pieces

    def format(self, delim='\t') -> str:
        cols = [str(self.idx), self.name, str(self.level), str(self.freq),
                ' '.join(str(k.idx) for k in self.kids) if self.kids else '']
        return delim.join(cols)

    @property
    def is_reserved(self) -> bool:
        return self.level == Level.reserved

    def signature(self) -> str:
        return f'{self.idx}:{self.name}→{"|".join(f"{k.idx}:{k.name}" for k in self.kids or [])}'

    @classmethod
    def parse(cls, line: str, vocab: List['Type'], delim='\t'):
        cols = line.strip().split(delim)
        idx, name, level, freq = cols[:4]
        idx, level, freq = int(idx), int(level), int(freq)
        kids = None
        if len(cols) > 4 and cols[4]:  # the last column maybe stripped out due to empty string
            kids = list(map(int, cols[4].split()))
            # kids shouldn't be referenced forward in index
            assert all(k < idx and k < len(vocab) for k in kids)
            kids = [vocab[k] for k in kids]
        return Type(name, level=level, idx=idx, freq=freq, kids=kids)

    def copy(self, **kwargs):
        assert kwargs
        # class is frozen, to update a field, make a copy and set new value
        args = {f.name: kwargs.pop(f.name, getattr(self, f.name)) for f in fields(self)}
        assert not kwargs, f'{kwargs} are unknown'
        return Type(**args)

    @classmethod
    def write_out(cls, table: List['Type'], out: Union[Path, str, TextIO]):

        if isinstance(out, Path) or isinstance(out, str):
            wrtr = open(out, 'w', encoding='utf8', errors='ignore')
        else:
            wrtr = out

        levels = dict(coll.Counter(v.level for v in table))
        max_level = max(levels.keys())
        meta = dict(total=len(table), version=__version__, levels=levels, max_level=max_level,
                    created=str(datetime.now()))
        meta = json.dumps(meta)
        wrtr.write(f"#{meta}\n")
        for i, item in enumerate(table):
            assert i == item.idx, f'{item} expected index {i}'
            wrtr.write(item.format() + '\n')
        if wrtr is not out:
            wrtr.close()
        log.info(f"Wrote {len(table)} to {wrtr.name}")

    @classmethod
    def read_vocab(cls, inp: Union[Path, str, TextIO]) -> Tuple[List['Type'], Optional[Dict]]:

        if isinstance(inp, Path) or isinstance(inp, str):
            rdr = open(inp, 'r', encoding='utf8', errors='ignore')
        else:
            rdr = inp

        lines = list(l.strip() for l in rdr)
        meta = None
        if lines[0].startswith("#{"):
            # metadata such as version; not used as of now
            meta = json.loads(lines.pop(0)[1:])

        # noinspection PyTypeChecker
        vocab: List[Type] = [None] * len(lines)
        for i, line in enumerate(lines):
            v = Type.parse(line=line.rstrip('\n'), vocab=vocab)
            assert v.idx == i
            vocab[i] = v
        if rdr is not inp:
            rdr.close()
        log.info(f"read {len(vocab)} types from {rdr.name}")
        return vocab, meta

    def get_permutations(self, name=False) -> List[List[int]]:
        """
        gets all possible permutations of kids that could compose thise
        :param name: get name of pieces instead of ids. default=False
        :return: List of permutations
        """
        perms = [[self.name if name else self.idx]]
        if self.kids:
            left, right = self.kids
            for left_kid in left.get_permutations(name=name):
                for right_kid in right.get_permutations(name=name):
                    perms.append(left_kid + right_kid)
        return perms

    def get_stochastic_split(self, name=False, split_ratio=0.1):
        if self.kids and random.random() < split_ratio:
            left, right = self.kids
            return left.get_stochastic_split(name=name, split_ratio=split_ratio) \
                   + right.get_stochastic_split(name=name, split_ratio=split_ratio)
        else:
            return [self.name if name else self.idx]


class EncoderScheme:

    def __init__(self, table: List[Type], validate=True, invertible=True):
        """

        :param table: list of `Type`s
        :param validate: validate that reserved types are found
        :param invertible: validate that the idx->str and str->idx are invertible
        """
        if validate:
            Reseved.validate(table)
            self.unk_idx = Reseved.UNK_IDX
        else:
            # at least UNK should be available
            assert table[Reseved.UNK_IDX].name == Reseved.UNK_TOK[0]
            # TODO: reverse lookup UNK IDX based on UNK_TOK name
            self.unk_idx = Reseved.UNK_IDX

        self.vocab_size = len(table)
        self.table = table
        self.idx_to_str = [t.name for t in table]
        if invertible:
            self.str_to_idx = {tok: idx for idx, tok in enumerate(self.idx_to_str)}
            assert len(self.idx_to_str) == len(self.str_to_idx)
        else:
            # keep the first occurrence TODO: maybe keep both and do random; str_to_idx be multiset
            self.str_to_idx = {}
            for idx, typ in enumerate(table):
                if typ.name in self.str_to_idx:
                    typ2 = table[self.str_to_idx[typ.name]]
                    log.debug(f"skip:: {typ.signature()}; it conflicts with {typ2.signature()}")
                else:
                    self.str_to_idx[typ.name] = idx
        self.invertible = invertible

    def __len__(self):
        return self.vocab_size


    @abc.abstractmethod
    def encode_str(cls, line: str) -> List[str]:
        raise NotImplementedError()

    @abc.abstractmethod
    def decode_str(cls, seq: List[str]) -> str:
        raise NotImplementedError()

    def encode(self, line: str) -> List[int]:
        pieces = self.encode_str(line)
        return [self.str_to_idx.get(piece, self.unk_idx) for piece in pieces]

    def decode(self, seq: List[int]) -> str:
        pieces = [self.idx_to_str[idx] for idx in seq]
        return self.decode_str(pieces)

    def encode_parallel(self, seqs: Iterator[str], n_cpus=N_CPUS) -> Iterator[List[int]]:
        return self.parallel_map(self.encode, seqs, n_cpus=n_cpus)

    @classmethod
    def parallel_map(cls, mapper, collection, n_cpus=N_CPUS, name='', chunksize=1000):
        assert n_cpus > 1, f'at least 2 CPUs needed. chunksize={chunksize}'
        log.info(f"Going to use {n_cpus} parallel processes {name}")
        with mp.Pool(processes=n_cpus) as pool:
            yield from pool.imap(mapper, collection, chunksize=chunksize)

    @classmethod
    @abc.abstractmethod
    def learn(cls, data: Iterator[str], **kwargs) -> List[Type]:
        raise NotImplementedError()

    @classmethod
    def get_init_vocab(cls, term_freqs, coverage: float = 0, line_count=None,
                       min_freq=WORD_MIN_FREQ,
                       vocab_size=-1):
        vocab = Reseved.with_reserved_types()
        res_stats = {r_type.name: term_freqs.pop(r_type.name) for r_type in vocab if
                     r_type.name in term_freqs}
        if res_stats:
            log.warning(f"Found reserved types in corpus: {res_stats}")
        # Order of trimming techs: 1. coverage, 2. min freqs, 3. size cut off
        unk_count = 0
        if coverage:
            assert 0 < coverage <= 1
            term_freqs, coverage_unk_count = filter_types_coverage(term_freqs, coverage=coverage)
            unk_count += coverage_unk_count
        term_freqs = sorted(term_freqs.items(), key=lambda x: x[1], reverse=True)
        if min_freq and min_freq > 1:
            log.info(f"Excluding terms with freq < {min_freq}; |freq >= 1|: {len(term_freqs):,}")
            unk_count += sum(f for t, f in term_freqs if f < min_freq)
            term_freqs = [(t, f) for t, f in term_freqs if f >= min_freq]
            log.info(f"|freq >= {min_freq}| : {len(term_freqs):,}")

        if vocab_size > 0 and len(vocab) + len(term_freqs) > vocab_size:
            log.info(f"Truncating vocab at size={vocab_size}")
            unk_count += sum(f for t, f in term_freqs[vocab_size - len(vocab):])
            term_freqs = term_freqs[:vocab_size - len(vocab)]

        # update reserved types with corpus freqs
        for idx, t in enumerate(vocab):
            freq = 0
            if t.name in res_stats:
                freq = res_stats.pop(t.name)
            if idx == Reseved.UNK_IDX:
                freq += unk_count
            if idx in {Reseved.BOS_IDX, Reseved.EOS_IDX, Reseved.CLS_IDX} and line_count:
                freq += line_count
            if freq:
                log.warning(f"Update frequency for reserved type {t} with {freq}")
                vocab[idx] = t.copy(freq=freq)
        vocab += [Type(name=name, idx=idx, freq=freq, level=cls.level)
                  for idx, (name, freq) in enumerate(term_freqs, start=len(vocab))]
        log.info(f"Total {cls} vocab size {len(vocab):,}")
        return vocab


class WordScheme(EncoderScheme):
    level = Level.word
    name = "word"

    @classmethod
    def encode_str(cls, line: str) -> List[str]:
        return line.split()

    @classmethod
    def decode_str(cls, seq: List[str]) -> str:
        return " ".join(seq)

    @classmethod
    def term_frequencies(cls, data: Iterator[str]) -> Tuple[Dict[str, int], int]:
        stats = coll.Counter()
        line_count = 0
        for line in tqdm(data, mininterval=1):
            stats.update(cls.encode_str(line.strip()))
            line_count += 1
        log.info(f"Found {len(stats):,} types and {sum(stats.values()):,} tokens")
        return stats, line_count

    @classmethod
    def read_term_freqs(cls, data: Iterator[str], delim='\t') -> Tuple[Dict[str, int], int]:
        stats = {}
        line_count = -1
        for idx, line in enumerate(data):
            line = line.rstrip('\n')
            if idx == 0 and line.startswith("#") and len(line.split(delim)) != 2:
                try:
                    import json
                    meta = json.loads(line[1:])  # skip # at index 0
                    line_count = meta.get('line_count', line_count)
                except:
                    pass
                continue
            term, freq = line.split("\t")
            stats[term.strip()] = int(freq)
        return stats, line_count

    @classmethod
    def learn(cls, data: Iterator[str], vocab_size: int = 0, min_freq: int = WORD_MIN_FREQ,
              coverage: float = 0, term_freqs=False, **kwargs) -> List[Type]:
        """
        :param data: input sentences
        :param vocab_size: max vocabulary size.
        :param min_freq: min frequency for inclusion in vocabulary. Excludes types with lower freq
        :param coverage: Character coverage
        :param term_freqs: is data the term_freqs ?
        :param kwargs: place holder for any extra args
        :return:
        """
        assert not kwargs, f'{kwargs} args are not allowed/understood'
        if term_freqs: # input is term_freqs
            log.info("Restoring term frequencies from input")
            stats, line_count = cls.read_term_freqs(data=data)
        else: # compute term freqs
            log.info("Computing term frequencies from raw data")
            stats, line_count = cls.term_frequencies(data=data)

        return cls.get_init_vocab(stats, coverage, line_count, min_freq, vocab_size)


class CharScheme(WordScheme):
    space_char = '▁'  # U+2581 same as google/sentencepiece
    level = Level.char
    name = "char"

    @classmethod
    def encode_str(cls, line: str) -> List[str]:
        return list(cls.space_char.join(line.split()))

    @classmethod
    def decode_str(cls, seq: List[str]) -> str:
        return ''.join(seq).replace(cls.space_char, ' ')

    @classmethod
    def learn(cls, data: Iterator[str], vocab_size: int = 0, min_freq: int = CHAR_MIN_FREQ,
              coverage: float = CHAR_COVERAGE, **kwargs) -> List[Type]:
        # learn() is same as parent class: WordScheme
        return super().learn(data, vocab_size, min_freq=min_freq, coverage=coverage, **kwargs)


class BPEScheme(CharScheme):
    level = Level.subword
    name = "bpe"

    def __init__(self, table: List[Type]):
        super().__init__(table=table, invertible=False)
        self.root = self.make_vocab_prefix_trie(self.table)
        log.info(f"Vocab size={len(self)}; trie root has nodes={self.root.size}"
                 f" but data_nodes={self.root.data_node_count}")
        assert self.unk_idx

    @classmethod
    def make_vocab_prefix_trie(cls, vocab: List[Type]):
        root: TrNode = TrNode[str, Type](idx='')
        for typ in vocab:
            node = root.get_node(idxs=typ.name, create_missing=True)
            node.name = typ.name
            node.data = typ
        assert not root.has_data  # root node is not data node
        return root


    def encode(self, line: str, split_ratio: float = 0.) -> List[int]:
        pieces = self.encode_str(line, split_ratio=split_ratio)
        return [self.str_to_idx.get(piece, self.unk_idx) for piece in pieces]

    def encode_str(self, line: str, split_ratio: float = 0.) -> List[str]:
        seq = self.space_char.join(line.strip().split()) + self.space_char
        res: List[int] = []
        """
          T  h  i  s  _  X  i  s  _ 
          T  h*  i  s  _* X i  s  _*     # * is data node
        ↑prev_node
        Say vocab has: Th, This_, is_, Xmas_, This_Xmas_ ;; 'X' is unk but Xmas_ is valid 
          T  h  i  s  _*  X  i  s  _*    # Greedy; This_* instead of Th*
          T  h  i  s  _*  ??  i  s  _*   # Try for This_Xmas_; not possible? Backup and Turn 'X' as unk  
          [This_, ??, is_]
        """
        # Note: could have been done recursively; but iterative is stack efficient
        # TODO: this can be improved; dont make too many sub string objs => iterate over char array
        data_node, data_idx = None, -1
        prev_node, idx = self.root, 0
        while seq and idx < len(seq) + 1:  # +1 extra loop to handle the buffer at the end
            if prev_node.has_data:
                data_node, data_idx = prev_node, idx

            if idx < len(seq) and seq[idx] in prev_node.kids:  # keep growing, i.e. advance
                prev_node = prev_node.kids[seq[idx]]
                idx += 1
            else:
                if data_node:  # the last seen data node exists
                    res.append(data_node.data.idx)
                    seq = seq[data_idx:]  # reset to the successors of last seen data node
                else:  # there was no data node seen; backup and make unk
                    res.append(self.unk_idx)  # unk the first char
                    seq = seq[1:]  # delete the first char from seq
                # either cases, reset the states; and 'seq' should be at least one unit smaller
                prev_node, idx = self.root, 0,
                data_node, data_idx = None, -1

        if split_ratio > 0:
            res = self.stochastic_split(res, split_ratio=split_ratio, name=False)

        return [self.table[idx].name for idx in res]

    def decode_str(self, seq: List[str]) -> str:
        return ''.join(seq).replace(self.space_char, ' ').strip()

    @classmethod
    def learn(cls, data: Iterator[str], vocab_size: int = 0, min_freq=WORD_MIN_FREQ,
              coverage=CHAR_COVERAGE, min_co_evidence=MIN_CO_EV, term_freqs=False, **kwargs) -> List[Type]:
        if min_co_evidence > 1 and vocab_size <= 1:
            log.info(f'Using Min-co-evidence={min_co_evidence} as stop condition. '
                     f'Treating vocab_size as maximum size')
            vocab_size = sys.maxsize
        assert vocab_size > 0
        assert not kwargs, f'{kwargs} args are not allowed/understood'
        if term_freqs:
            log.info("Reading term freqs from input")
            tfs, line_count = WordScheme.read_term_freqs(data)
        else:
            log.info("Computing term freqs from input")
            tfs, line_count = WordScheme.term_frequencies(data)

        def init_vocab_factory(char_types):
            return CharScheme.get_init_vocab(char_types, line_count=line_count,
                                             coverage=coverage, min_freq=1)

        from .bpe import BPELearn
        vocab = BPELearn.learn_subwords(term_freqs=tfs, vocab_size=vocab_size,
                                        init_vocab_factory=init_vocab_factory,
                                        min_co_evidence=min_co_evidence)
        return vocab

    def stochastic_split(self, seq, split_ratio, name=False):
        res = []
        for idx in seq:
            res += self.table[idx].get_stochastic_split(name=name, split_ratio=split_ratio)
        return res


#########################
REGISTRY = {
    'char': CharScheme,
    'word': WordScheme,
    'bpe': BPEScheme,
    'subword': BPEScheme
}


def learn_vocab(inp, level, model, vocab_size, min_freq=1, term_freqs=False,
                char_coverage=CHAR_COVERAGE, min_co_ev=MIN_CO_EV):
    if not min_freq or min_freq < 1:
        min_freq = WORD_MIN_FREQ if level == 'word' else CHAR_MIN_FREQ
        log.info(f"level={level} => default min_freq={min_freq}")
    else:
        log.info(f"level={level} => user given min_freq={min_freq}")
    log.info(f"Learn Vocab for level={level} and store at {model}")
    log.info(f"data ={inp}")
    Scheme = REGISTRY[level]
    args = {}
    if level != 'word':
        args['coverage'] = char_coverage  # no char_coverage for word
    if level == 'bpe':
        args['min_co_evidence'] = min_co_ev
    table = Scheme.learn(inp, vocab_size=vocab_size, min_freq=min_freq, term_freqs=term_freqs,
                         **args)
    Type.write_out(table=table, out=model)


def load_scheme(path: Union[str, Path, TextIO]) -> EncoderScheme:
    types, meta = Type.read_vocab(path)
    assert meta
    max_level = meta['max_level']
    levels = [CharScheme, BPEScheme, WordScheme]
    assert max_level < len(levels)
    Scheme = levels[max_level]
    return Scheme(table=types)


def encode(inp: Iterator[str], scheme: EncoderScheme, indices=False) \
        -> Iterator[Union[List[str], List[int]]]:
    for line in inp:
        if indices:
            yield scheme.encode(line)
        else:
            yield scheme.encode_str(line)


def decode(inp: Iterator[str], scheme: EncoderScheme, indices=False) -> Iterator[str]:
    for line in inp:
        seq = line.split()
        if indices:
            seq = [int(x) for x in seq]
            line = scheme.decode(seq)
        else:
            line = scheme.decode_str(seq)
        yield line



#!/usr/bin/env python
#
# Author: Thamme Gowda [tg (at) isi (dot) edu] 
# Created: 2019-10-25


import argparse
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Iterator, List, Any

import numpy as np
from tqdm import tqdm

from nlcodec import EncoderScheme, log, load_scheme

NDArray = np.ndarray


@dataclass()
class CodecQuality:
    n_types: int
    imbalance: Any
    n_tokens: int
    mean_seq_len: float
    unk_count: int = 0


class QualityEstimator:

    def __init__(self, codec: EncoderScheme):
        self.codec = codec

    def estimate(self, lines: Iterator[str]) -> CodecQuality:
        seqs = (self.codec.encode(line) for line in lines)
        return self._estimate(seqs)

    @staticmethod
    def validate_distribution(distr):
        assert distr.dtype == np.float
        assert abs(np.sum(distr) - 1.0) < 1e-4  # sum to 1
        assert np.sum(distr < 0) == 0  # no negatives

    @classmethod
    def earth_mov_dist(cls, before: NDArray, after: NDArray, moving_cost=0.5, validate=True):
        if validate:
            cls.validate_distribution(before)
            cls.validate_distribution(after)
        assert len(before) == len(after)
        assert moving_cost > 0
        return  moving_cost * np.sum(np.abs(after - before))

    @classmethod
    def EMD_simple(cls, stats: NDArray, moving_cost=1):
        distr = stats / np.sum(stats)
        k = len(stats)
        assert moving_cost > 0
        return 0.5 * moving_cost * np.sum(np.abs(distr - 1/k))

    @classmethod
    def kl_div(self, base, distr):
        """
        Dkl (P || Q) should be read such that P is base from which Q's  divergence is computed
        Dkl (P || Q) =  ∑ P(x) log(P(x)/Q(x)) = - ∑ P(x) log(Q(x) / P(x))
        if base p is 0, then log(p/q) is not meaningful.
        if q is 0, we are in trouble; either log(0) = ∞ or log(1 / 0) which is division-by-zero
        so make sure q is never zero. Use JS Div which solves this
        :param base: P in DKL(P || Q)
        :param distr: Q in DKL(P || Q)
        :return: DKL(P || Q)
        """
        total = 0.0
        for px, qx in zip(base, distr):
            if px > 0:
                assert qx > 0
                total += px * np.log(px / qx)
        return total

    @classmethod
    def js_div(cls, base: NDArray, distr: NDArray):
        cls.validate_distribution(base)
        cls.validate_distribution(distr)
        assert len(base) == len(distr)

        middle = 0.5 * (base + distr)
        return 0.5 * (cls.kl_div(base=base, distr=middle) + cls.kl_div(base=distr, distr=middle))

    def _estimate(self, seqs: Iterator[List[int]]) -> CodecQuality:
        stats = [0] * self.codec.vocab_size
        # lengths = coll.defaultdict(int)
        n_seqs = 0
        for seq in tqdm(seqs):
            n_seqs += 1
            for tok in seq:
                stats[tok] += 1

        n_toks = sum(stats)
        mean_len = n_toks / n_seqs
        n_types = self.codec.vocab_size

        unk_count = stats[self.codec.unk_idx]

        # exclude the reserved types which dont surface in prediction
        #excludes = [typ for typ in self.codec.table if typ.is_reserved and stats[typ.idx] == 0]
        #n_types -= len(excludes)  # exclude reserved
        #for typ in reversed(excludes):  # delete from right
        #    log.info(f'Remove type from balance check: {typ}')
        #    del stats[typ.idx]
        n_toks = sum(stats)
        imb_measure = self.EMD_simple(stats=stats)
        return CodecQuality(n_types=n_types, imbalance=imb_measure, n_tokens=n_toks,
                            mean_seq_len=mean_len, unk_count=unk_count)


def estimate(codec_path, data):
    codec = load_scheme(codec_path)
    estimator = QualityEstimator(codec)
    estimation = estimator.estimate(data)
    print(estimation)


def parse_args():
    p = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    p.add_argument('-c', '--codec', dest='codec_path', type=Path,
                   required=True, help='Path to codec vocabulary')
    p.add_argument('-d', '--data', type=argparse.FileType('r'), default=sys.stdin,
                   help='Input data file path')
    return vars(p.parse_args())


def main():
    args = parse_args()
    estimate(**args)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
#
# Author: Thamme Gowda [tg (at) isi (dot) edu] 
# Created: 2019-10-25

from typing import List, Dict, Union, Optional, TypeVar, Generic, Any
import heapq


class LnNode:  # doubly linked list node data structure; used for learning BPE
    __slots__ = 'val', 'left', 'right', 'freq', 'data'

    def __init__(self, val: int, left: Optional['LnNode'] = None,
                 right: Optional['LnNode'] = None, freq: int = 1, data: Any = None):
        self.val = val
        self.left = left
        self.right = right
        self.freq = freq
        self.data = data

    def __eq__(self, other):
        # caution: these calls are recursive on left and right; cycles would cause infinite loop
        return (other.val == self.val and other.freq == self.freq and
                other.left == self.left and other.right == self.right)

    def __hash__(self):
        return id(self)  # quick and dirty hash; not sure how this mess if we use multiprocessing

    def delete(self, unlink=True):
        """
        deletes this node from the list
        :return:
        """
        x, y = self.left, self.right
        if x:  # right links : x → self → y  => x → y
            x.right = y
        if y:  # left links  : x ← self ← y  => x ← y
            y.left = x
        if unlink:
            self.left = self.right = None

    @property
    def is_unlinked(self):
        return self.right is None and self.left is None

    @classmethod
    def from_seq(cls, string: Union[str, List[int]], freq=1, data=None) -> List['LnNode']:
        """
        makes a doubly linked list from string
        :param string: input string (of integers recommended)
        :param freq: frequency of string in corpus (for scaling
        :return: List of Nodes, doubly linked to lefts and rights
        """
        nodes = [cls(ch, freq=freq, data=data) for ch in string]
        for i, n in enumerate(nodes):
            if i > 0:
                n.left = nodes[i - 1]
            if i + 1 < len(nodes):
                n.right = nodes[i + 1]
        return nodes

    def __repr__(self):
        lefts, rights = [], []

        cur = self.left
        while cur:
            lefts.append(str(cur.val))
            cur = cur.left
        cur = self.right
        while cur:
            rights.append(str(cur.val))
            cur = cur.right
        return ' '.join(reversed(lefts)) + f' *{self.val}* ' + ' '.join(rights)


T = TypeVar('T')  # T for index; Dont use I, since 'I' agree with pep008
D = TypeVar('D')  # D for data


class TrNode(Generic[T, D]):  # Trie Node or Tree Node

    __slots__ =  'idx', 'name', 'data', 'parent', 'kids'

    def __init__(self, idx: T, name: Optional[str] = None, data: Optional[D] = None,
                 parent: Optional['TrNode'] = None, kids: Dict[T, 'TrNode'] = None):
        self.idx = idx
        self.name = name
        self.data = data
        self.parent = parent
        self.kids = dict() if kids is None else kids

    def get_node(self, idxs, create_missing: bool = True) -> 'TrNode[T, D]':
        if not idxs:
            return self
        if create_missing and idxs[0] not in self.kids:  # make one
            self.kids[idxs[0]] = TrNode(idx=idxs[0], parent=self)
        return self.kids[idxs[0]].get_node(idxs=idxs[1:], create_missing=create_missing)

    @property
    def n_kids(self):
        """Number of immediate children"""
        return len(self.kids)

    @property
    def has_data(self):
        return self.data is not None

    @property
    def size(self):
        """Number of nodes in the tree at this node. Counts self and all the kids"""
        return 1 + sum(k.size for k in self.kids.values())

    @property
    def data_node_count(self):
        return (1 if self.has_data else 0) + sum(k.data_node_count for k in self.kids.values())


class MaxHeap:
    "Offers Max Heap which is a wrapper to fast min heap implementation from heapq "

    def __init__(self, items: Dict[Any, int]):
        self.min_heap = [(-val, node) for node, val in items.items()]
        heapq.heapify(self.min_heap)

    def push(self, node, val):
        """
        Adds a node to heap while maintaining heap property
        :param node: node data
        :param val: priority value
        :return:
        """
        heapq.heappush(self.min_heap, (-val, node))

    def pop(self):
        """returns the max item (task, priority)"""
        val, task = heapq.heappop(self.min_heap)
        return task, -val

    def peek(self):
        val, task = self.min_heap[0]
        return task, -val

    def __len__(self):
        return len(self.min_heap)



#!/usr/bin/env python
#
# Author: Thamme Gowda [tg (at) isi (dot) edu] 
# Created: 7/19/20


import itertools
import json
import math
import os
import pickle
import random
import shutil
from collections import namedtuple, defaultdict
from pathlib import Path
from typing import List, Iterator, Dict, Any, Tuple

import numpy as np

from nlcodec import log
from nlcodec.utils import as_path

Array = np.ndarray
Record = Tuple[Array]

DEF_TYPE = np.uint16  # uint16 is [0, 65,535]
DEF_MIN = np.iinfo(DEF_TYPE).min
DEF_MAX = np.iinfo(DEF_TYPE).max
DEF_PART_SIZE = 5_000_000
DEF_MAX_PARTS = 1_000


def best_dtype(mn, mx):
    """
    determines best (integer) data type for a given range of integer
    :param mn: min value
    :param mx: max value
    :return: numpy data type
    """
    assert isinstance(mn, (int, np.integer))
    assert isinstance(mx, (int, np.integer))
    # smaller to bigger
    options = [np.uint8, np.int8, np.uint16, np.int16, np.uint32, np.int32, np.uint64, np.int64]
    found_type = None
    for int_type in options:
        info = np.iinfo(int_type)
        if info.min <= mn and mx <= info.max:  # completely fits inside
            found_type = int_type
            break
    assert found_type, f'Could not find an integet type for [{min},{max}]'
    return found_type


def part_path_pads(n_parts: int):
    return math.ceil(math.log10(n_parts))


class SeqField:
    """A field of unequal length sequences"""

    class Builder:
        """
        Writer to assist creation of
        """
        # 100MB; each int takes 28 bytes + 8 byte for ref
        BUFFER_SIZE = 100_000_000 / (28 + 8)

        def __init__(self, name, buf_size=None):

            self.name = name
            self.ids = {}
            self.frozen = np.array([], dtype=np.int8)   # start with smallest type
            self.buffer = []
            self.refs = []
            self.max_len = 0
            self.buf_size = buf_size or self.BUFFER_SIZE

        def append(self, id, arr):
            assert len(arr) == 0 or isinstance(arr[0], (int, np.integer))
            self.ids[id] = len(self.refs)
            self.max_len = max(len(arr), self.max_len)
            self.refs.append([len(self.frozen) + len(self.buffer), len(arr)])    # start, len
            self.buffer.extend(arr)
            if len(self.buffer) >= self.buf_size:
                self.shrink()
            return self

        def shrink(self):
            if self.buffer:
                dtype = best_dtype(mn=min(self.buffer), mx=max(self.buffer))
                data = np.array(self.buffer, dtype=dtype)
                self.frozen = np.concatenate((self.frozen, data))
                self.buffer = []

        def build(self):
            self.shrink()
            assert not self.buffer
            refs_type = best_dtype(mn=0, mx=max(len(self.frozen), self.max_len))
            return SeqField(name=self.name, ids=self.ids,
                            refs=np.array(self.refs, dtype=refs_type), data=self.frozen)

    def __init__(self, name: str, ids: Dict[Any, int], refs: Array, data: Array):
        self.name = name
        self.ids = ids
        self.refs = refs
        self.data = data
        assert len(data.shape) == 1
        assert refs.shape == (len(ids), 2)
        assert refs[-1][0] <= len(data)

    def __len__(self):
        return len(self.ids)

    def __getitem__(self, item):
        start, length = self.refs[self.ids[item]]
        return self.data[start: start + length]

    def get_len(self, _id):
        return self.refs[self.ids[_id]][1]

    def keys(self):
        return self.ids.keys()

    def values(self):
        for start, _len in self.refs:
            yield self.data[start: start + _len]

    def items(self):
        for id, idx in self.ids.items():
            start, _len = self.refs[idx]
            yield self.data[start: start + _len]

    def lengths(self):
        return ((id, self.refs[idx, 1]) for id, idx in self.ids.items())

    @classmethod
    def create(cls, name, recs) -> 'SeqField':
        builder = cls.Builder(name=name)
        for id, arr in recs:
            builder.append(id, arr)
        return builder.build()

    @classmethod
    def create_many(cls, names: List[str], recs: Iterator[List[List[int]]]) -> List['SeqField']:
        builders = [cls.Builder(name) for name in names]
        for id, rec in recs:
            assert len(rec) == len(builders)
            for b, col in zip(builders, rec):
                b.append(id, col)
        return [b.build() for b in builders]


class Db:

    def __init__(self, fields: List[SeqField], rec_type=None):
        assert all(isinstance(f, SeqField) for f in fields)
        self.field_names = [f.name for f in fields]
        self._rec_type = rec_type
        self.fields = {f.name: f for f in fields}
        self._len = len(fields[0])
        self.ids = set(fields[0].ids.keys())
        for fn, fd in self.fields.items():
            assert self._len == len(fd)  # all have same num of recs
            assert self.ids == set(fd.ids.keys())  # and same IDs

    # noinspection PyPep8Naming
    def RecType(self):
        if not self._rec_type:
            self._rec_type = namedtuple('RecType', ['id'] + self.field_names)
        return self._rec_type

    def __len__(self):
        return self._len

    def __getstate__(self):
        state = self.__dict__.copy()
        state['_rec_type'] = None  # bcoz it is not pickleable
        return state

    def save(self, path):
        log.debug(f"Saving to {path}")
        with open(path, 'wb') as f:
            pickle.dump(self, f)

    @classmethod
    def load(cls, path, rec_type=None) -> 'Db':
        log.debug(f"Loading from {path}")
        with open(path, 'rb') as f:
            obj = pickle.load(f)
        assert isinstance(obj, cls)
        if rec_type:
            obj._rec_type = rec_type
        return obj

    @classmethod
    def create(cls, recs, field_names, has_id=False, path=None):
        """
        :param recs: Iterator[List[List[int]]] or Iterator[(id, List[List[int]])]
        :param field_names: field names in records
        :param has_id: set True if recs already have id ie. tuple(id, fields),
           when set to False (default) ids are auto generated
        :param path: path to save on disk (optional)
        :return:
        """
        if not has_id:
            recs = enumerate(recs)
        fields = SeqField.create_many(field_names, recs)
        db = cls(fields=fields)
        if path:
            db.save(path)
        return db

    def __getitem__(self, _id):
        cols = tuple(self.fields[fn][_id] for fn in self.field_names)
        return self.RecType()(_id, *cols)

    def __iter__(self):
        for _id in self.ids:
            yield self[_id]

    def _make_eq_len_batch_ids(self, max_toks, max_sents, min_len=1):
        fields = list(self.fields.values())
        rows = []
        skip_counter = defaultdict(int)
        for _id in self.ids:
            lens = tuple(field.get_len(_id) for field in fields)
            if min(lens) < min_len:
                skip_counter[f'len < {min_len}'] += 1
            else:
                rows.append((_id, max(lens)))
        if len(skip_counter) > 0:
            log.warning(f"Skipped :: {skip_counter}")
        rows = np.array(rows)  # id, len
        np.random.shuffle(rows)  # in-place, along the first axis; for extra rand within len group
        rows = rows[rows[:, 1].argsort()]  # sort by second col wiz len
        batches = []
        batch = []
        max_len = 0
        for _id, _len in rows:
            if _len < 1:
                log.warning(f"Skipping record {_id}, either source or target is empty")
                continue

            if (len(batch) + 1) * max(max_len, _len) > max_toks or len(batch) > max_sents:
                if _len > max_toks:
                    raise Exception(f'Unable to make a batch of {max_toks} toks'
                                    f' with a seq of len:{_len}')
                batches.append(np.array(batch))
                batch = []  # new batch
                max_len = 0

            batch.append(_id)  # this one can go in
            max_len = max(max_len, _len)
        if batch:
            batches.append(np.array(batch))
        return batches

    def make_eq_len_ran_batches(self, max_toks, max_sents=float('inf')):

        batches = self._make_eq_len_batch_ids(max_toks=max_toks, max_sents=max_sents)
        if not batches:
            raise Exception(f'Found no data. Please check config data paths')
        log.info(f"length sorted random batches = {len(batches)}. Shuffling🔀...")
        # every pass introduce some randomness
        random.shuffle(batches)

        for batch_ids in batches:
            batch = [self[_id] for _id in batch_ids]
            yield batch


class MultipartDb:

    @classmethod
    def slices(cls, stream, size):
        _stream = iter(stream)
        try:
            while True:
                segment = itertools.islice(_stream, size)
                n = next(segment)  # so that it raises StopIteration when empty, to break the loop
                yield itertools.chain([n], segment)
        except StopIteration:
            pass

    @classmethod
    def create(cls, path, recs, field_names, has_id=False, overwrite=False,
               part_size=DEF_PART_SIZE, max_parts=DEF_MAX_PARTS):
        if not has_id:
            recs = enumerate(recs)
        builder = cls.Writer(path=path, field_names=field_names, overwrite=overwrite,
                             max_parts=max_parts)

        part_num = -1
        for sliced in cls.slices(recs, part_size):
            part_num += 1
            builder(part_num, recs=sliced)
        builder.close()
        return cls.load(path=path)

    @classmethod
    def load(cls, path, rec_type=None, keep_in_mem=False) -> 'MultipartDb':
        path = as_path(path)
        assert path.is_dir()
        flag_file = path / '_SUCCESS'
        assert flag_file.exists(), 'Looks like the save was incomplete or corrupted'
        meta = json.loads(flag_file.read_text())
        part_files = []
        rec_counts = []
        for part_name, stats in meta['parts'].items():
            part_file = path / part_name
            part_file.exists()
            part_files.append(part_file)
            rec_counts.append(stats['count'])
        field_names = meta['field_names']
        rec_type = rec_type or namedtuple('RecType', field_names)
        assert len(part_files) > 0
        return cls(parts=part_files, rec_counts=rec_counts, rec_type=rec_type,
                   keep_in_mem=keep_in_mem)

    def __init__(self, parts: List[Path], rec_counts: List[int], rec_type, keep_in_mem=False):
        self.part_paths = parts
        self.rec_counts = rec_counts
        self._len = sum(rec_counts)
        self.rec_type = rec_type
        self.keep_in_mem = keep_in_mem or len(parts) == 1  # if its only one part, just keep it
        self.mem = [None] * len(parts)
        if keep_in_mem:
            self.mem = [Db.load(p, rec_type=rec_type) for p in parts]

    def __len__(self):
        return self._len

    def __iter__(self):
        for idx, path in enumerate(self.part_paths):
            if self.keep_in_mem:
                part = self.mem[idx]
            else:
                part = Db.load(path)
            yield from part

    def make_eq_len_ran_batches(self, max_toks, max_sents=float('inf'), join_ratio=0.0) -> Iterator[List]:
        # shuffle the parts
        buff = list(zip(self.part_paths, self.mem))
        random.shuffle(buff)

        for path, part in buff:
            if part is None:
                part = Db.load(path, rec_type=self.rec_type)
            yield from part.make_eq_len_ran_batches(max_toks=max_toks, max_sents=max_sents)

    class Writer:

        def __init__(self, path, field_names: List[str], overwrite=False,
                     max_parts=DEF_MAX_PARTS):
            self.field_names = field_names
            path = as_path(path)
            if path.exists() and len(os.listdir(path)) > 0:
                if overwrite:
                    log.warning(f"Removing existing data at {path}")
                    shutil.rmtree(path)
                else:
                    raise Exception(f'{path} already exists. not overwriting it')
            path.mkdir(parents=True, exist_ok=True)
            self.path = path
            self.part_path_pad = part_path_pads(max_parts)

        def __call__(self, part_num: int, recs):
            # assume recs have ids created externally
            part = Db.create(recs, field_names=self.field_names, has_id=True)
            part_path = self.path / f'part-{part_num:0{self.part_path_pad}d}'
            part.save(part_path)
            meta_path = part_path.with_suffix('.meta')
            meta = dict(count=len(part), size=part_path.stat().st_size)
            meta = json.dumps(meta, ensure_ascii=False, indent=2)
            meta_path.write_text(meta)
            return part_path, len(part)

        def close(self):
            parts = {}
            for meta_path in self.path.glob("part-*.meta"):
                part_name = meta_path.name.rstrip('.meta')
                parts[part_name] = json.loads(meta_path.read_text())
            meta = dict(parts=parts, field_names=self.field_names)
            flag_file = self.path / '_SUCCESS'
            flag_file.write_text(json.dumps(meta, indent=2))

        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc_val, exc_tb):
            if exc_tb is None:
                self.close()
                return True
            else:
                return False



#!/usr/bin/env python
#
# Author: Thamme Gowda [tg (at) isi (dot) edu] 
# Created: 6/13/20
import json
import logging as log
from pathlib import Path
from typing import List, Dict, Tuple

from nlcodec import spark as spark_util
from nlcodec import __version__,  __epilog__

log.basicConfig(level=log.INFO)



def word_counts(paths: List[Path], dedup=True, spark=None) \
        -> Tuple[Dict[str, int], Dict[str, int], int]:
    """

    :param paths: List of paths of plain text files
    :param dedup: deduplicate lines
    :param spark: (optional) provide a spark instance;
     By default, a local spark session is launched and stopped
    :return: word_freqs, char_freqs, line_count
    """
    own_spark = not spark
    spark = spark or spark_util.get_session()
    log.info(f"Spark Session is up; you may check web UI")
    try:
        for p in paths:
            p.exists()
        dfs = [spark.read.text(str(p)) for p in paths]
        df = dfs[0]
        if len(dfs) > 1:
            for df2 in dfs[1:]:
                df = df.union(df2)
        # Deduplicate lines
        if dedup:
            log.info("Dropping duplicates")
            df = df.drop_duplicates()
        SENT_TAG = '<!S!>'
        # Tokenize
        # also include a special token, one per line to count lines
        # expand words as sep records
        word_freqs = df.rdd.flatMap(lambda row: [SENT_TAG] + row.value.strip().split()) \
            .map(lambda word: (word, 1)) \
            .reduceByKey(lambda v1, v2: v1 + v2) \
            .collectAsMap()
        line_count = word_freqs.pop(SENT_TAG)
        # word_freqs = list(sorted(word_freqs.items(), key=lambda x: x[1], reverse=True))

        # char freqs
        char_freqs = (spark.sparkContext.parallelize(word_freqs.items())
                      .flatMap(lambda r: [(ch, r[1]) for ch in r[0]])
                      .reduceByKey(lambda a, b: a + b)
                      .collectAsMap())

        return word_freqs, char_freqs, line_count

    finally:
        # only stop the own spark, not the user given spark
        if own_spark:
            log.info(f"Stopping spark session")
            spark.stop()


def write_stats(stats: Dict[str, int], out, **meta):
    stats = sorted(stats.items(), key=lambda x: x[1], reverse=True)
    stats = list(stats)
    tot = sum(f for t, f in stats)

    header = dict(types=len(stats), tokens=tot)
    if meta:
        header.update(meta)
    header = json.dumps(header)
    out.write(f"#{header}\n")
    for term, freq in stats:
        out.write(f'{term}\t{freq}\n')
    log.info(f"Wrote {out}")


def main(args=None):
    args = args or parse_args()
    word_freqs, char_freqs, line_count = word_counts(args.inp, dedup=args.dedup)
    if args.word_freqs:
        write_stats(stats=word_freqs, out=args.word_freqs, line_count=line_count)
    if args.char_freqs:
        write_stats(stats=char_freqs, out=args.char_freqs, line_count=line_count)


def parse_args():
    import argparse
    p = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                epilog=__epilog__)
    p.add_argument('-v', '--version', action='version', version=f'%(prog)s {__version__}')
    p.add_argument('-i', '--inp', type=Path, nargs='+',
                   help='Input file paths')
    p.add_argument('-wf', '--word_freqs', type=argparse.FileType('w'), default=None,
                   help='Output file path for word frequencies')
    p.add_argument('-cf', '--char_freqs', type=argparse.FileType('w'), default=None,
                   help='Output file path for character frequencies')

    p.add_argument('-dd', '--dedup', action='store_true',
                   default=True, help='Deduplicate the sentences: use only unique sequences')
    p.add_argument('-ndd', '--no-dedup', dest='dedup', action='store_false', default=False,
                   help='Do not deduplicate. ')
    args = p.parse_args()
    assert args.word_freqs or args.char_freqs, \
        'At least one of --word_freqs --char_freqs paths be given to write the stats'
    return args


if __name__ == '__main__':
    main()



#!/usr/bin/env python
#
# map-reduce using multiprocessing
#
# Author: Thamme Gowda [tg (at) isi (dot) edu] 
# Created: 5/2/20
import collections as coll
import multiprocessing as mp
from dataclasses import dataclass
from typing import Dict, Iterator, Any, Callable, TypeVar, Generic
import itertools
import logging as log
from functools import partial

log.basicConfig(level=log.DEBUG)

@dataclass
class SingletonException(Exception):
    item: Any


def pair_subsequent(iterator: Iterator, default=None, throw_singleton=False):
    """
    groups subsequent items in to pairs
    :param iterator: stream from which pairs are to be made
    :param default: default value to be used if there is an odd number of items in iterator
    :return:
    """
    last = None
    idx = 0
    for idx, item in enumerate(iterator, start=1):
        if idx % 2 == 0:  # we have a pair
            yield last, item
        last = item
    if throw_singleton:
        # at least one item should be there
        assert idx > 0, 'At least one item was expected but the input was empty'
        if idx == 1:
            raise SingletonException(last)
    if idx % 2 == 1:  # the last item un paired
        yield last, default


T = TypeVar('T')


@dataclass
class IterativeReducer(Generic[T]):
    func: Callable[[T, T], T]  # the reduce function (T, T) -> T
    default: T  # default value is a pairing item for the odd numbered item

    def dispatcher(self, args):
        assert len(args) == 2
        return self.func(args[0], args[1])

    def __call__(self, iterator: Iterator[T], pool=map) -> T:
        try:
            while True:  # break by exception when  all are reduced to singleton
                pairs = pair_subsequent(iterator=iterator, default=self.default,
                                        throw_singleton=True)
                iterator = pool(self.dispatcher, pairs)
                # iterator isn't evaluated since it is lazy. So, the code goes into infinite loop
                # the trick is to evaluate one item at least! aaha moment
                # This is black magic !
                iterator = itertools.chain([next(iterator)], iterator)
        except SingletonException as result:
            # reduced all the way to a singleton element
            return result.item


def ireduce(reduce_func: Callable[[T, T], T], default: T, data: Iterator[T], pool=map):
    ireducer = IterativeReducer(func=reduce_func, default=default)
    return ireducer(data, pool=pool)


@dataclass
class MapReduce:

    mapper: Callable
    reducer: Callable
    default: Any  # default for reducer, to pair the odd item

    def __call__(self, input, map_pool=None, reduce_pool=None):
        mapped = map_pool(self.mapper, input)
        reduced = ireduce(reduce_func=self.reducer, default=self.default, data=mapped,
                          pool=reduce_pool)
        return reduced

    def multiprocess(self, input, map_procs=2, red_procs=2, chunksize=1000):
        """
        runs map reduce on multi processes
        :param input: iterator of input that goes to mapper function
        :param map_procs: number of mapper processes
        :param red_procs: number of reduce processes
        :param chunksize: chunksize for localising data to processes.
         a smaller value consumes less memory but slow, higher value consumes more memory but fast.
        :return: final result from reducer
        """
        # TODO: support process local reduction
        assert map_procs >= 1
        assert red_procs >= 1
        log.info(f"map_processes: {map_procs}, reduce_processes:{red_procs}")

        with mp.get_context("spawn").Pool(processes=map_procs) as map_pool, \
                mp.get_context("spawn").Pool(processes=red_procs) as red_pool:
            map_pooler = partial(map_pool.imap, chunksize=chunksize)
            red_pooler = partial(red_pool.imap, chunksize=chunksize)
            return self(input, map_pool=map_pooler, reduce_pool=red_pooler)


def word_count_line(text):
    """Mapper task : maps line of text into term freqs at the sentence level"""
    return coll.Counter(text.split())


def merge_freqs(tfs1: Dict[str, int], tfs2: Dict[str, int]) -> Dict[str, int]:
    """
    Merges term-frequencies from tfs1 and tfs2 dictionaries.
    This is an inplace operation i.e. It updates the largest dictionary in the argument inplace
    and returns that as result.
    """
    # keep the largest, dump the smallest;
    keep, dump = tfs1, tfs2
    if len(keep) < len(dump):
        keep, dump = dump, keep
    for k, v in dump.items():
        keep[k] = keep.get(k, 0) + v
    return keep


def WordCounter() -> MapReduce:
    return MapReduce(mapper=word_count_line, reducer=merge_freqs, default={})

def main(lines):
    tfs =  WordCounter().multiprocess(lines)
    tfs = sorted(tfs.items(), key=lambda x:x[1], reverse=True)
    for t,f in tfs:
        print(f'{t}\t{f:,}')

if __name__ == '__main__':
    lines = ["hello this is a line 1", "and this is a line 2", "and blah blah"]
    import sys
    #lines = sys.stdin
    main(lines=lines)



#!/usr/bin/env python
#
# Author: Thamme Gowda [tg (at) isi (dot) edu]
# Created: 2019-11-12

import argparse
import collections as coll
import logging as log
import sys
from pathlib import Path
from typing import Iterator, Set, Dict, Tuple

from tqdm import tqdm

from nlcodec import load_scheme, WordScheme

log.basicConfig(level=log.INFO)
TermFreqs = Dict[str, int]


def term_freqs(inp : Iterator[str]) -> TermFreqs:
    tfs = coll.Counter()

    for line in tqdm(inp, mininterval=1, dynamic_ncols=True):
        toks = line.strip().split()
        tfs.update(toks)
    return  tfs

def partition_vocab_toks(inp : Iterator[str], vocab: Set[str]) -> Tuple[TermFreqs, TermFreqs]:
    tfs = term_freqs(inp)
    iv_stats = {t:f for t, f in tfs.items() if t in vocab}
    oov_stats = {t: f for t, f in tfs.items() if t not in vocab}
    return  iv_stats, oov_stats

def main(inp, model_path):
    codec = load_scheme(model_path)
    assert isinstance(codec, WordScheme)
    vocab = set(codec.str_to_idx.keys())
    iv_toks, oov_toks = partition_vocab_toks(inp, vocab)
    n_iv_types = len(iv_toks)
    n_iv_toks = sum(iv_toks.values())
    n_oov_types = len(oov_toks)
    n_oov_toks = sum(oov_toks.values())
    total_types = n_iv_types + n_oov_types
    total_toks = n_iv_toks + n_oov_toks

    print("*\tInVocab\tOOV")
    print(f"Types\t{n_iv_types}\t{n_oov_types}")
    print(f"Token Count\t{n_iv_toks}\t{n_oov_toks}")
    print(f"Token %\t{100*n_iv_toks/total_toks:.2f}\t{100*n_oov_toks/total_toks:.2f}")


if __name__ == '__main__':
    p = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                description="Compute term frequencies for ")
    p.add_argument('-i', '--inp', type=argparse.FileType('r'), default=sys.stdin,
                   help='Input file path')
    p.add_argument('-m', '--model', dest='model_path', type=Path, required=True,
                   help="model aka vocabulary file")
    args = vars(p.parse_args())
    main(**args)



#!/usr/bin/env python
#
# Author: Thamme Gowda [tg (at) isi (dot) edu] 
# Created: 7/31/20

import logging as log
from pathlib import Path
from nlcodec import load_scheme
from nlcodec import spark as spark_util
from nlcodec import utils

from nlcodec import __version__,  __epilog__

log.basicConfig(level=log.INFO)


def main(args=None):
    args = args or parse_args()
    src_codec = load_scheme(args.src_model)
    tgt_codec = src_codec
    if args.tgt_model:
        tgt_codec = load_scheme(args.tgt_model)

    with spark_util.session() as session:
        df = spark_util.read_raw_bitext_tok(
            spark=session, src_path=args.src_text, tgt_path=args.tgt_text, truncate=args.truncate,
            src_len=args.src_len, tgt_len=args.tgt_len,
            src_tokenizer=src_codec.encode, tgt_tokenizer=tgt_codec.encode)
        rdd = df.rdd.map(lambda r: (r.id, (r.x, r.y)))
        db = spark_util.rdd_as_db(rdd=rdd, db_path=args.db_path, field_names=('x', 'y'),
                                  repartition=args.num_parts)
        log.info(f"stored {len(db)} recs at {args.db_path}")


# noinspection PyTypeChecker
def parse_args():
    import argparse
    p = argparse.ArgumentParser(description="Encodes bitext nlcdec and stores them in db, "
                                            " using pyspark backend.",
                                formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                epilog=__epilog__)
    p.add_argument('-v', '--version', action='version', version=f'%(prog)s {__version__}')
    p.add_argument("-sm", "--src-model", metavar='PATH', type=Path, required=True, help="source model")
    p.add_argument("-tm", "--tgt-model", metavar='PATH',type=Path, required=False,
                   help="target model; default=same as src-model")
    p.add_argument("-st", "--src-text", metavar='PATH', type=Path, required=True, help="source text")
    p.add_argument("-tt", "--tgt-text", metavar='PATH', type=Path, required=True, help="target text")

    p.add_argument("-db", "--db", metavar='PATH', dest='db_path', type=Path, required=True,
                   help="Path to store db")
    p.add_argument("-np", "--num-parts", metavar='N', type=int, default=0,
                   help="A value greater than 0 forces the db to have these many parts")

    p.add_argument("-sn", "--src-col", type=str, default='x', help="source column name")
    p.add_argument("-tn", "--tgt-col", type=str, default='y', help="target column name")

    p.add_argument("-sl", "--src-len", type=int, default=256, help="source length max")
    p.add_argument("-tl", "--tgt-len", type=int, default=256, help="target length max")
    utils.add_bool_arg(p, 'truncate', default=True, help='truncate longer sentences',
                       nhelp="drop longer sentences")

    p.add_argument("-spark", "--spark_master", default='local[*]', help="Use Spark master")
    p.add_argument("-dm", "--driver_mem", default='4g', help="Memory for spark driver")

    args = p.parse_args()
    assert args.src_text.exists()
    assert args.tgt_text.exists()
    return args


if __name__ == '__main__':
    main()



#!/usr/bin/env python
#
# Author: Thamme Gowda [tg (at) isi (dot) edu] 
# Created: 2019-10-25

__version__ = '0.3.1'
__description__ = """nlcodec is a collection of encoding schemes for natural language sequences. 
nlcodec.db is a efficient storage and retrieval layer for integer sequences of varying lengths."""
PROJECT_HOME = 'https://github.com/isi-nlp/nlcodec'

from pathlib import Path
__epilog__ = f'Visit https://github.com/isi-nlp/nlcodec or https://isi-nlp.github.io/nlcodec/' \
             f' to learn more. You\'re currently using version {__version__} loaded ' \
             f' from {Path(__file__).parent}'


DEF_MIN_CO_EV = 95   # recommended by Gowda and May (2020)
DEF_WORD_MIN_FREQ = 1  # minimum times a word should exist to be used for word vocab
DEF_CHAR_MIN_FREQ = 20  # minimum times a char should be seen to be included in init vocab
DEF_CHAR_COVERAGE = 0.9995  # Credits to google/sentencepiece for this idea;


import logging as log
log.basicConfig(level=log.INFO)

from nlcodec.codec import (EncoderScheme, WordScheme, CharScheme, BPEScheme, Type, Reseved, REGISTRY,
                    learn_vocab, load_scheme, Level, encode, decode)
from nlcodec.dstruct import LnNode, TrNode, MaxHeap



#!/usr/bin/env python
#
# Author: Thamme Gowda [tg (at) isi (dot) edu] 
# Created: 7/31/20

from pathlib import Path
from typing import Union, Tuple, Dict
import os
from contextlib import contextmanager

from nlcodec import log
from pyspark.sql import DataFrame, SparkSession
from pyspark.sql.types import StructType, StructField, LongType
from nlcodec.db.core import MultipartDb


def get_session(app_name="NLCodec", master=None, driver_mem=None, config: Dict[str, str] = None):
    """
    gets Spark Session
    :param app_name: name Spark App
    :param master: spark master (optional)
    :param driver_mem: driver memory (optional)
    :param config: extra config
    :return:
    """
    driver_mem = driver_mem or os.environ.get("SPARK_DRIVER_MEM", "4g")
    builder = (SparkSession.builder.appName(app_name)
               .master(master or os.environ.get("SPARK_MASTER", "local[*]"))
               .config("spark.driver.memory", driver_mem))
    if config:
        for k, v in config.items():
            builder.config(k, v)

    spark = builder.getOrCreate()
    return spark


@contextmanager
def session(*args, **kwargs):
    spark = get_session(*args, **kwargs)
    try:
        yield spark
    finally:
        spark.stop()


def read_raw_bitext(spark, src_file: Union[str, Path], tgt_file: Union[str, Path],
                    src_name='src_raw', tgt_name='tgt_raw') -> Tuple[DataFrame, int]:
    """
    reads bitext to a dataframe
    :param spark:  spark session
    :param src_file: source file to read from
    :param tgt_file:  target file to read from
    :param src_name: name for source col in DF
    :param tgt_name: name for target col in DF
    :return: DataFrame
    """
    if not isinstance(src_file, str):
        src_file = str(src_file)
    if not isinstance(tgt_file, str):
        tgt_file = str(tgt_file)

    src_df = spark.read.text(src_file).withColumnRenamed('value', src_name)
    tgt_df = spark.read.text(tgt_file).withColumnRenamed('value', tgt_name)

    n_src, n_tgt = src_df.count(), tgt_df.count()
    assert n_src == n_tgt, f'{n_src} == {n_tgt} ?'
    log.info(f"Found {n_src:,} parallel records in {src_file, tgt_file}")

    def with_idx(sdf):
        new_schema = StructType(sdf.schema.fields + [StructField("idx", LongType(), False), ])
        return sdf.rdd.zipWithIndex().map(lambda row: row[0] + (row[1],)).toDF(
            schema=new_schema)

    src_df = with_idx(src_df)
    tgt_df = with_idx(tgt_df)
    bitext_df = src_df.join(tgt_df, 'idx', "inner")
    # n_bitext = bitext_df.count()
    # assert n_bitext == n_src, f'{n_bitext} == {n_src} ??'
    return bitext_df, n_src


def read_raw_bitext_tok(spark, src_path: Union[str, Path], tgt_path: Union[str, Path],
                        truncate: bool, src_len: int, tgt_len: int, src_tokenizer,
                        tgt_tokenizer) -> DataFrame:
    """
    Reads raw text and applies tokenizer in parallel
    :param spark:  SparkSession
    :param src_path: source path
    :param tgt_path: target path
    :param truncate: truncate or skip longer sequences ?
    :param src_len:  max source length
    :param tgt_len:  max target length
    :param src_tokenizer: source tokenizer
    :param tgt_tokenizer: target tokenize
    :return:  DataFrame
    """
    raw_df, n_recs = read_raw_bitext(spark, src_path, tgt_path, src_name='x', tgt_name='y')
    tok_rdd = (raw_df.rdd
               .filter(lambda r: r.x and r.y)  # exclude None
               .map(lambda r: (r.idx, src_tokenizer(r.x), tgt_tokenizer(r.y)))
               .filter(lambda r: len(r[1]) and len(r[2]))
               # exclude empty, if tokenizer created any
               )
    if truncate:
        tok_rdd = tok_rdd.map(lambda r: (r[0], r[1][:src_len], r[2][:tgt_len]))
    else:
        tok_rdd = tok_rdd.filter(lambda r: len(r[1]) < src_len and len(r[2]) < tgt_len)

    # dataframes doesnt support numpy arrays, so we cast them to python list
    a_row = tok_rdd.take(1)[0]
    if not isinstance(a_row[1], list):
        # looks like np NDArray or torch tensor
        tok_rdd = tok_rdd.map(lambda r: (r[0], r[1].tolist(), r[2].tolist()))
    tok_rdd.map(lambda r: (r[0],))
    df = tok_rdd.toDF(['id', 'x', 'y'])
    return df


def rdd_as_db(rdd, db_path: Path, field_names=('x', 'y'), repartition=0, **kwargs) -> MultipartDb:
    # input format: each rec in rdd should be : (id, (x, y)
    if repartition and repartition > 0:
        rdd = rdd.repartition(repartition)

    with MultipartDb.Writer(db_path, field_names=field_names, **kwargs) as writer:
        n = rdd.mapPartitionsWithIndex(writer).count()  # parallel write
    db = MultipartDb.load(db_path)
    log.info(f"Wrote {len(db)} recs in {len(db.part_paths)} parts at {db_path}")
    return db



#!/usr/bin/env python
#
# tool to estimate how the mean sequence length shrink with bpe merges
# Author: Thamme Gowda [tg (at) isi (dot) edu] 
# Created: 2019-11-28
import argparse
import collections as coll
import sys
from pathlib import Path
from typing import Iterator, Dict, Set, Optional, Union

import numpy as np
from tqdm import tqdm

from nlcodec import log, BPEScheme, Type, Level
from nlcodec.bpe import Bigram, LnNode
from nlcodec.utils import max_RSS

NDArray = np.ndarray


class BPECodecXt(BPEScheme):
    """
    Use for merging one by one for the sake of tracking
    """

    def __init__(self, path: Union[str,Path]):
        types, meta = Type.read_vocab(path)
        assert meta
        assert meta['max_level'] == Level.subword
        init_table = [t for t in types if t.level < Level.subword]
        self.merge_buffer = [t for t in types if t.level >= Level.subword]
        for m in self.merge_buffer:
            assert len(m.kids) == 2  # exactly two pieces

        super().__init__(table=init_table) # start with character
        self.root = self.make_vocab_prefix_trie(self.table)
        assert self.unk_idx
        log.info(f"initial table has {len(init_table)}; buffer has {len(self.merge_buffer)}")

    def merges_remaining(self):
        return len(self.merge_buffer)

    def __len__(self):
        return len(self.table)

    def total_types(self):
        return len(self) + self.merges_remaining()

    def peek_merge(self) -> Type:
        return self.merge_buffer[0]

    def merge(self) -> Type:
        assert self.merge_buffer
        merge = self.merge_buffer.pop(0)
        assert len(self.table) == len(self.idx_to_str) == merge.idx
        self.table.append(merge)
        self.idx_to_str.append(merge.name)
        self.str_to_idx[merge.name] = merge.idx
        node = self.root.get_node(idxs=merge.name, create_missing=True)
        node.name = merge.name
        node.data = merge
        return merge


class BpeTracker:
    """
    Tracker for tracking how lengths and imbalance change over BPE merges
    """
    def __init__(self, bpe: BPECodecXt, data: Iterator[str]):
        self.bpe = bpe
        self.uni: NDArray = np.zeros(bpe.total_types(), dtype=np.long) # term freq ; unigrams

        # Bigram to sequence references
        self.bi_ixs: Dict[Bigram, Set[LnNode]] = coll.defaultdict(set)
        self.seq_lens: Optional[NDArray] = None
        self.create_index(data)

    def create_index(self, data):
        """
        1. encoder all : List[str] -> List[List[Int]]
        2. Count unigram freqs => uni
              2a) measure imbalance
              2b) measure seq lens
        3. Index all bigrams:: bi_ix (a,b) -> Node 
           Oh, yeah, convert all List[Int] to Doubly Linked List
           Create seq_len: List[int] each node should have an integer index of sequence in corpus
           
        4. ask bpe for next merge. gives (a,b) -> t
            a) count how many bigrams in index. 
                insert uni[t] = count
                reduce  uni[a], uni[b] by that count
            b) loop through each indexed node 
               remove b. update a;  update links :: x a b y => x t y
               insert bi_ix[x, t] and bi_ix[t, y]
               using seq number stored on node, decrement length by 1. assert len >= 1
            c) measure imbalance, and avg length
        """
        log.info("Encoding and creating index")
        enc_data  = self.bpe.encode_parallel(data)
        uni: Dict[int, int] = coll.defaultdict(int)
        bi_ixs: Dict[Bigram, Set[LnNode]] = coll.defaultdict(set)
        seq_lens = []
        with tqdm(enumerate(enc_data)) as data_bar:
            for idx, seq in data_bar:
                seq = LnNode.from_seq(seq, data=idx)
                seq_lens.append(len(seq))
                for node in seq:
                    uni[node.val] += 1
                for node in seq[:-1]:
                    bi_ixs[(node.val, node.right.val)].add(node)
                bar_msg = f'MaxRSS={max_RSS()[1]}'
                data_bar.set_postfix_str(bar_msg, refresh=False)

        self.seq_lens = np.array(seq_lens)
        log.info(f"Found {len(self.seq_lens)} sentences")
        for idx, freq in uni.items():
            self.uni[idx] = freq
        self.bi_ixs = bi_ixs

    @classmethod
    def earth_mov_dist(cls, stats: NDArray):
        k = len(stats)
        total = np.sum(stats)
        probs = stats / total
        diffs = np.abs(probs -  (1/k))
        return  0.5 * np.sum(diffs)

    def mean_seq_len(self):
        return np.mean(self.seq_lens)

    def track_merges(self, do_log=False):
        uni = self.uni
        bi_ixs = self.bi_ixs
        seq_lens = self.seq_lens
        res = []
        while self.bpe.merges_remaining() > 0:
            """
            ask bpe for next merge. gives (a,b) -> t
            a) count how many bigrams in index. 
                insert uni[t] = count
                reduce  uni[a], uni[b] by that count
            b) loop through each indexed node 
               remove b. update a;  update links :: x a b y => x t y
               insert bi_ix[x, t] and bi_ix[t, y]
               using seq number stored on node, decrement length by 1. assert len >= 1
            c) measure imbalance, and avg length
            """
            k = len(self.bpe)
            emd = self.earth_mov_dist(self.uni[:k])
            mu_len = self.mean_seq_len()
            res.append((k, emd, mu_len))
            if do_log:
                log.info(f"{k}: imbalance={emd:g} mean_seq_len={mu_len:g}."
                         f"Next: {self.bpe.peek_merge().signature()}")
            new_type = self.bpe.merge()
            a_type, b_type = new_type.kids
            a, b = a_type.idx, b_type.idx
            update_nodes = bi_ixs.pop((a, b), None)
            if update_nodes:
                freq = sum(u.freq for u in update_nodes)
                uni[new_type.idx] = freq
                uni[a] -= freq
                uni[b] -= freq
                for u in update_nodes:
                    b_node = u.right
                    if u.is_unlinked or (a == b and new_type.idx in (u.val, b_node.val)):
                        # happens due to repeats like X A A A A Y
                        uni[new_type.idx] -= u.freq
                        uni[a] += u.freq
                        uni[b] -= u.freq
                        continue
                    assert u.val == a
                    assert b_node.val == b, f'expected {a, b} found {u.val, b_node.val} at {b_node}'
                    u.val = new_type.idx
                    b_node.delete(unlink=True)

                    seq_idx = u.data
                    seq_lens[seq_idx] -= 1
                    assert seq_lens[seq_idx] > 0

                    if u.left:
                        if  bi_ixs.get((u.left.val, a)):
                            bi_ixs[(u.left.val, a)].remove(u.left)
                        bi_ixs[(u.left.val, u.val)].add(u.left)
                    if u.right:
                        if bi_ixs.get((b, u.right.val)):
                            bi_ixs[(b, u.right.val)].remove(b_node)
                        bi_ixs[(u.val, u.right.val)].add(u)
        k = len(self.bpe)
        emd = self.earth_mov_dist(self.uni[:k])
        mu_len = self.mean_seq_len()
        res.append((k, emd, mu_len))
        log.info(f"{k}: imbalance={emd:g} mean_seq_len={mu_len:g}.")
        return res


def estimate(codec_path, data):
    bpe = BPECodecXt(codec_path)
    tracker = BpeTracker(bpe, data)
    tracker.track_merges(do_log=True)

def parse_args():
    p = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    p.add_argument('-c', '--codec', dest='codec_path', type=Path,
                   required=True, help='Path to codec vocabulary')
    p.add_argument('-d', '--data', type=argparse.FileType('r'), default=sys.stdin,
                   help='Input data file path')
    return vars(p.parse_args())


def main():
    args = parse_args()
    estimate(**args)


if __name__ == '__main__':
    main()


#!/usr/bin/env python
#
# Author: Thamme Gowda [tg (at) isi (dot) edu] 
# Created: 7/29/20
import math
from dataclasses import dataclass
from pathlib import Path
from typing import List, Iterator, Iterable, Optional, Tuple, Union

import numpy as np

from nlcodec import log
from .core import Db, MultipartDb

Array = np.ndarray


class IdExample:
    """
    An object of this class holds an example in sequence to sequence dataset
    """
    __slots__ = ('x', 'y', 'id', 'x_raw', 'y_raw')

    def __init__(self, id, x, y):
        self.x: Array = x
        self.y: Array = y
        self.id = id
        self.x_raw: Optional[str] = None
        self.y_raw: Optional[str] = None

    def val_exists_at(self, side, pos: int, exist: bool, val: int):
        assert side == 'x' or side == 'y'
        assert pos == 0 or pos == -1
        seq = self.x if side == 'x' else self.y
        if exist:
            if seq[pos] != val:
                if pos == 0:
                    seq = np.append(np.int32(val), seq)
                else:  # pos = -1
                    seq = np.append(seq, np.int32(val))
                # update
                if side == 'x':
                    self.x = seq
                else:
                    self.y = seq
        else:  # should not have val at pos
            assert seq[pos] != val

    def eos_check(self, side, exist):
        raise

    def __getitem__(self, key):
        if key == 'x_len':
            return len(self.x)
        elif key == 'y_len':
            return len(self.y)
        else:
            return getattr(self, key)


@dataclass
class BatchMeta:
    pad_idx: int
    bos_idx: int
    eos_idx: int
    add_bos_x: bool = False
    add_bos_y: bool = False
    add_eos_x: bool = True
    add_eos_y: bool = True


class Batch:
    """
    An object of this class holds a batch of examples
    """
    _x_attrs = ('x_len', 'x_seqs')
    _y_attrs = ('y_len', 'y_seqs')

    def __init__(self, batch: List[IdExample], sort_dec=False, batch_first=True,
                 meta: BatchMeta = None):
        """
        :param batch: List fo Examples
        :param sort_dec: True if the examples be sorted as descending order of their source sequence lengths
        :Param Batch_First: first dimension is batch
        """
        assert isinstance(meta, BatchMeta)
        self.meta = meta
        self.batch_first = batch_first

        self.bos_eos_check(batch, 'x', meta.add_bos_x, meta.add_eos_x)
        if sort_dec:
            batch = sorted(batch, key=lambda _: len(_.x), reverse=True)
        self._len = len(batch)
        self.x_len = np.array([len(e.x) for e in batch])
        self.x_toks = np.sum(self.x_len)
        self.max_x_len = np.max(self.x_len)

        self.x_seqs = np.full(shape=(self._len, self.max_x_len), fill_value=self.meta.pad_idx,
                              dtype=np.long)
        for i, ex in enumerate(batch):
            self.x_seqs[i, :len(ex.x)] = ex.x
        if not batch_first:  # transpose
            self.x_seqs = np.transpose(self.x_seqs)
        self.x_raw = None
        if batch[0].x_raw:
            self.x_raw = [ex.x_raw for ex in batch]

        first_y = batch[0].y
        self.has_y = first_y is not None
        if self.has_y:
            self.bos_eos_check(batch, 'y', meta.add_bos_y, meta.add_eos_y)
            self.y_len = np.array([len(e.y) for e in batch])
            self.y_toks = np.sum(self.y_len)
            self.max_y_len = np.max(self.y_len)
            self.y_seqs = np.full(shape=(self._len, self.max_y_len), fill_value=meta.pad_idx,
                                  dtype=np.long)
            for i, ex in enumerate(batch):
                self.y_seqs[i, :len(ex.y)] = ex.y

            if not batch_first:  # transpose
                self.y_seqs = np.transpose(self.y_seqs)
            self.y_raw = None
            if batch[0].y_raw:
                self.y_raw = [ex.y_raw for ex in batch]

    @staticmethod
    def val_exists(obj, field_name, pos: int, exist: bool, val: int):
        assert pos == 0 or pos == -1
        seq = getattr(obj, field_name)
        if exist:
            if seq[pos] != val:
                if pos == 0:
                    seq = np.append(np.int32(val), seq)
                else:  # pos = -1
                    seq = np.append(seq, np.int32(val))
                # update
                setattr(obj, field_name, seq)
        else:  # should not have val at pos
            assert seq[pos] != val

    def bos_eos_check(self, batch: List[IdExample], side: str, bos: bool, eos: bool):
        """
        ensures and inserts (if needed) EOS and BOS tokens
        :param batch:
        :param side: which side? choices: {'x', 'y'}
        :param bos: True if should have BOS, False if should not have BOS
        :param eos: True if should have EOS, False if should not have EOS
        :return: None, all modifications are inplace of batch
        """
        for ex in batch:
            self.val_exists(ex, field_name=side, pos=0, exist=bos, val=self.meta.bos_idx)
            self.val_exists(ex, field_name=side, pos=-1, exist=eos, val=self.meta.eos_idx)

    def __len__(self):
        return self._len

    def to(self, device):
        """Move this batch to given device"""
        for name in self._x_attrs + (self._y_attrs if self.has_y else []):
            setattr(self, name, getattr(self, name).to(device))
        return self

    def make_autoreg_mask(self, tgt):
        "Create a mask to hide padding and future words for autoregressive generation."
        return self.make_autogres_mask_(tgt, self.meta.pad_idx)

    @staticmethod
    def make_autogres_mask_(seqs, pad_val: int):
        "Create a mask to hide padding and future words."
        tgt_mask = (seqs != pad_val).unsqueeze(1)
        tgt_mask = tgt_mask & subsequent_mask(seqs.size(-1)).type_as(tgt_mask.data)
        return tgt_mask


def subsequent_mask(size):
    """
    Mask out subsequent positions. upper diagonal elements should be zero
    :param size:
    :return: mask where positions are filled with zero for subsequent positions
    """
    # upper diagonal elements are 1s, lower diagonal and the main diagonal are zeroed
    ones = np.ones((size, size), dtype=np.int8)
    triu = np.triu(ones, k=1)
    # invert it
    mask = triu == 0
    mask = mask.unsqueeze(0)
    return mask


class BatchIterable(Iterable[Batch]):

    # This should have been called as Dataset
    def __init__(self, data_path: Path, batch_size: Union[int, Tuple[int, int]], batch_meta: BatchMeta,
                 sort_desc: bool = False, batch_first: bool = True, sort_by: str = None,
                 keep_in_mem=False):
        """
        Iterator for reading training data in batches
        :param data_path: path to TSV file
        :param batch_size: number of tokens on the target size per batch; or (max_toks, max_sents)
        :param sort_desc: should the mini batch be sorted by sequence len (useful for RNN api)
        :param keep_in_mem: keep all parts in memory for multipartdb;
           for single part, of course, the part remains in memory.
        """
        self.batch_meta = batch_meta
        self.sort_desc = sort_desc
        if isinstance(batch_size, int):
            self.max_toks, self.max_sents = batch_size, float('inf')
        else:
            self.max_toks, self.max_sents = batch_size
        self.batch_first = batch_first
        self.sort_by = sort_by
        self.data_path = data_path
        self.keep_in_mem = keep_in_mem
        assert sort_by in (None, 'eq_len_rand_batch', 'random')
        if not isinstance(data_path, Path):
            data_path = Path(data_path)

        assert data_path.exists(), f'Invalid State: {data_path} is NOT found.'
        if data_path.is_file():
            self.data = Db.load(data_path, rec_type=IdExample)
        elif self.data_path.is_dir():
            self.data = MultipartDb.load(data_path, rec_type=IdExample, keep_in_mem=keep_in_mem)
        else:
            raise Exception(f'Invalid State: {data_path} is should be a file or dir.')

        log.info(f'Batch Size = {batch_size}, sort_by={sort_by}')

    def read_all(self):
        batch = []
        max_len = 0
        for ex in self.data:
            if min(len(ex.x), len(ex.y)) == 0:
                log.warning("Skipping a record,  either source or target is empty")
                continue

            this_len = max(len(ex.x), len(ex.y))
            if (len(batch) + 1) * max(max_len, this_len) <= self.max_toks and len(batch) < self.max_sents :
                batch.append(ex)  # this one can go in
                max_len = max(max_len, this_len)
            else:
                if this_len > self.max_toks:
                    raise Exception(f'Unable to make a batch of {self.max_toks} toks'
                                    f' with a seq of x_len:{len(ex.x)} y_len:{len(ex.y)}')
                # yield the current batch
                yield Batch(batch, sort_dec=self.sort_desc, batch_first=self.batch_first,
                            meta=self.batch_meta)
                batch = [ex]  # new batch
                max_len = this_len
        if batch:
            log.debug(f"\nLast batch, size={len(batch)}")
            yield Batch(batch, sort_dec=self.sort_desc, batch_first=self.batch_first,
                        meta=self.batch_meta)

    def make_eq_len_ran_batches(self):
        batches = self.data.make_eq_len_ran_batches(max_toks=self.max_toks, max_sents=self.max_sents)
        for batch in batches:
            yield Batch(batch, sort_dec=self.sort_desc, batch_first=self.batch_first,
                        meta=self.batch_meta)

    def __iter__(self) -> Iterator[Batch]:
        if self.sort_by == 'eq_len_rand_batch':
            yield from self.make_eq_len_ran_batches()
        else:
            yield from self.read_all()

    @property
    def num_items(self) -> int:
        return len(self.data)

    @property
    def num_batches(self) -> int:
        return int(math.ceil(len(self.data) / self.max_toks))



#!/usr/bin/env python
#
# Author: Thamme Gowda [tg (at) isi (dot) edu] 
# Created: 7/31/20

import logging as log
from typing import Dict, Any
from pathlib import Path
from nlcodec import DEF_WORD_MIN_FREQ, DEF_CHAR_MIN_FREQ, DEF_CHAR_COVERAGE, DEF_MIN_CO_EV
from nlcodec import __version__, learn_vocab
from nlcodec.term_freq import word_counts, write_stats
from nlcodec import utils
from nlcodec import spark
from nlcodec import __version__, __epilog__

log.basicConfig(level=log.INFO)

__description__ = 'learn BPE on pyspark'


def parse_args() -> Dict[str, Any]:
    import argparse
    # noinspection PyTypeChecker
    p = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                prog='nlcodec-learn',
                                description=__description__,
                                epilog=__epilog__)
    p.add_argument('-v', '--version', action='version', version=f'%(prog)s {__version__}')

    p.add_argument('-i', '--inp', type=Path, nargs='+',
                   help='Paths to all the input files from which to extract terms.')
    p.add_argument('-m', '--model', type=Path, help='Path to model aka vocabulary file',
                   required=True)

    p.add_argument('-vs', '--vocab-size', type=int, default=-1,
                   help='Vocabulary size. This is required for'
                        ' "bpe", but optional for "word" and "char" models, specifying it'
                        ' will trim the vocabulary at given top most frequent types.')
    p.add_argument('-l', '--level', choices=['char', 'word', 'bpe'], default='bpe',
                   help='Encoding Level')
    p.add_argument('-mf', '--min-freq', default=None, type=int,
                   help='Minimum frequency of types for considering inclusion in vocabulary. '
                        'Types fewer than this frequency will be ignored. '
                        f'For --level=word or --level=bpe, freq is type freq and '
                        f' default is {DEF_WORD_MIN_FREQ}.'
                        f'for --level=char, characters fewer than this value'
                        f' will be excluded. default={DEF_CHAR_MIN_FREQ}')

    p.add_argument('-cv', '--char-coverage', default=DEF_CHAR_COVERAGE, type=float,
                   help='Character coverage for --level=char or --level=bpe')

    p.add_argument('-mce', '--min-co-ev', default=DEF_MIN_CO_EV, type=int,
                   help='Minimum Co-evidence for BPE merge. Valid when --level=bpe')

    p.add_argument('-spark', '--spark-master', default="local[*]", help="Spark master")
    p.add_argument('-dm', '--driver-mem', default="4g", help="Spark driver memory")
    utils.add_bool_arg(p, name='dedup', default=False,
                       help='Deduplicate the sentences: use only unique sequences')

    args = vars(p.parse_args())
    return args


def main():
    args = parse_args()
    model_path: Path = args['model']
    words_file = model_path.with_suffix(".wordfreq.gz")
    chars_file = model_path.with_suffix(".charfreq.gz")
    stats_file = chars_file if args['level'] == 'char' else words_file

    with utils.log_resources(name="extract stats"):
        master = args.pop('spark_master')
        driver_mem = args.pop('driver_mem')
        dedup = args.pop('dedup')
        if stats_file.exists():
            log.warning(f"{stats_file} exists, reusing it. please delete it if this is wrong.")
        else:
            inp = args.pop('inp')
            with spark.session(master=master, driver_mem=driver_mem) as session:
                words, chars, line_count = word_counts(paths=inp, dedup=dedup, spark=session)
            with utils.IO.writer(words_file) as out:
                write_stats(words, out, line_count=line_count)
            with utils.IO.writer(chars_file) as out:
                write_stats(chars, out, line_count=line_count)

    assert stats_file.exists()
    with utils.log_resources(name=f"learning {args['level']} vocab"):
        with utils.IO.reader(stats_file) as inp:
            learn_vocab(inp=inp, term_freqs=True, **args)


if __name__ == '__main__':
    main()



import nlcodec
from pathlib import Path

from setuptools import setup, find_packages, Extension

long_description = Path('README.md').read_text(encoding='utf-8', errors='ignore')

classifiers = [  # copied from https://pypi.org/classifiers/
    'Development Status :: 5 - Production/Stable',
    'Intended Audience :: Developers',
    'Topic :: Utilities',
    'Topic :: Text Processing',
    'Topic :: Text Processing :: General',
    'Topic :: Text Processing :: Filters',
    'Topic :: Text Processing :: Linguistic',
    'License :: OSI Approved :: Apache Software License',
    'Programming Language :: Python :: 3',
]

setup(
    name='nlcodec',
    version=nlcodec.__version__,
    description=nlcodec.__description__,
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='Apache Software License 2.0',
    classifiers=classifiers,
    python_requires='>=3.7',
    url='https://github.com/thammegowda/bpepp',
    download_url='https://github.com/thammegowda/bpepp',
    platforms=['any'],
    author='Thamme Gowda',
    author_email='tgowdan@gmail.com',

    packages=find_packages(exclude=['experimental']),
    entry_points={
        'console_scripts': [
            'nlcodec=nlcodec.__main__:main',
            'nlcodec-learn=nlcodec.learn:main',
            'nlcodec-db=nlcodec.bitextdb:main',
            'nlcodec-freq=nlcodec.term_freq:main'
        ],
    },
    install_requires=['tqdm', 'numpy']
)



#!/usr/bin/env python
#
# Author: Thamme Gowda [tg (at) isi (dot) edu] 
# Created: 2019-11-12
import resource
import sys
from pathlib import Path
from typing import List, Any, Iterable, Dict, Tuple, Union
import collections as coll
from nlcodec import log
from tqdm import tqdm
import gzip
import time
from contextlib import contextmanager
from datetime import timedelta


def make_n_grams(sent: List[Any], n):
    assert n > 0
    return [tuple(sent[i: i + n]) for i in range(len(sent) - n + 1)]


def make_n_grams_all(sents: Iterable[List[Any]], n):
    grams = coll.Counter()
    n_sent = 0
    for sent in tqdm(sents, mininterval=1, dynamic_ncols=True):
        grams.update(make_n_grams(sent, n))
        n_sent += 1
    log.info(f"Made {n}-grams: types={len(grams)}; tokens={sum(grams.values())}")
    return grams


def filter_types_coverage(types: Dict[str, int], coverage=1.0) -> Tuple[Dict[str, int], int]:
    assert  0 < coverage <= 1
    tot = sum(types.values())
    includes = {}
    cum = 0
    types  = sorted(types.items(), key=lambda x: x[1], reverse=True)
    for t, f in types:
        cum += f / tot
        includes[t] = f
        if cum >= coverage:
            break
    log.info(f'Coverage={cum:g}; requested={coverage:g}')
    excludes = {ch: ct for ch, ct in types if ch not in includes}
    unk_count = sum(excludes.values())
    log.warning(f'UNKed total toks:{unk_count} types={len(excludes)} from types:{excludes}')
    return includes, unk_count


def as_path(path: Union[str, Path]) -> Path:
    """
    returns an instance of Path, optionally converting string to Path when needed
    :param path: instance of str or Path
    :return: instance of Path
    """
    if not isinstance(path, Path):
        path = Path(path)
    return path


def max_RSS(who=resource.RUSAGE_SELF) -> Tuple[int, str]:
    """Gets memory usage of current process, maximum so far.
    Maximum so far, since the system call API doesnt provide "current"
    :returns (int, str)
       int is a value from getrusage().ru_maxrss
       str is human friendly value (best attempt to add right units)
    """
    mem = resource.getrusage(who).ru_maxrss
    h_mem = mem
    if 'darwin' in sys.platform:  # "man getrusage 2" says we get bytes
        h_mem /= 10 ** 3  # bytes to kilo
    unit = 'KB'
    if h_mem >= 10 ** 3:
        h_mem /= 10 ** 3  # kilo to mega
        unit = 'MB'
    return mem, f'{int(h_mem)}{unit}'


class IO:
    """File opener and automatic closer"""

    def __init__(self, path, mode='r', encoding=None, errors=None):
        self.path = path if type(path) is Path else Path(path)
        self.mode = mode
        self.fd = None
        self.encoding = encoding if encoding else 'utf-8' if 't' in mode else None
        self.errors = errors if errors else 'replace'

    def __enter__(self):

        if self.path.name.endswith(".gz"):  # gzip mode
            self.fd = gzip.open(self.path, self.mode, encoding=self.encoding, errors=self.errors)
        else:
            if 'b' in self.mode:  # binary mode doesnt take encoding or errors
                self.fd = self.path.open(self.mode)
            else:
                self.fd = self.path.open(self.mode, encoding=self.encoding, errors=self.errors,
                                         newline='\n')
        return self.fd

    def __exit__(self, _type, value, traceback):
        self.fd.close()

    @classmethod
    def reader(cls, path, text=True):
        return cls(path, 'rt' if text else 'rb')

    @classmethod
    def writer(cls, path, text=True, append=False):
        return cls(path, ('a' if append else 'w') + ('t' if text else 'b'))


@contextmanager
def log_resources(name=""):
    """
    logs time and memory utilized by a code block
    :param name: some name to identify code block
    :return:
    """
    st = time.time()
    st_mem = max_RSS()[1]
    try:
        yield name
    finally:
        delta = timedelta(seconds=time.time() - st)
        et_mem = max_RSS()[1]
        log.info(f"{name} Time: {delta}; Mem: {st_mem} --> {et_mem}")


def add_bool_arg(parser, name, default=False, help=None, nhelp=None):
    group = parser.add_mutually_exclusive_group(required=False)
    group.add_argument(f'--{name}', dest=name, action='store_true', help=help)
    group.add_argument(f'--no-{name}', dest=name, action='store_false', help=nhelp or f'See --{name}')
    parser.set_defaults(**{name: default})
